#include "AnyRichEdit.h"


INT AnyRichEdit_ExGetSelPos(HWND hEd, INT* pnStartPos, INT* pnEndPos)
{
  CHARRANGE cr;

  SendMessage(hEd, EM_EXGETSEL, 0, (LPARAM) &cr);
  if (pnStartPos != NULL)
    *pnStartPos = (INT) cr.cpMin;
  if (pnEndPos != NULL)
    *pnEndPos = (INT) cr.cpMax;

  return cr.cpMin;
}

void AnyRichEdit_ExLimitText(HWND hEd, INT nMaxLength)
{
  SendMessage(hEd, EM_EXLIMITTEXT, 0, (LPARAM) nMaxLength);
}

INT AnyRichEdit_ExLineFromChar(HWND hEd, INT nCharacterIndex)
{
  return ((INT) SendMessage(hEd, EM_EXLINEFROMCHAR, 0, (LPARAM) nCharacterIndex));
}

void AnyRichEdit_ExSetSelPos(HWND hEd, INT nStartPos, INT nEndPos)
{
  CHARRANGE cr;
  cr.cpMin = (LONG) nStartPos;
  cr.cpMax = (LONG) nEndPos;
  SendMessage(hEd, EM_EXSETSEL, 0, (LPARAM) &cr);
}

INT AnyRichEdit_FirstVisibleCharIndex(HWND hEd)
{
  RECT editRect;
  if (AnyRichEdit_GetEditRect(hEd, &editRect))
  {
    POINTL pt;
    pt.x = editRect.left;
    pt.y = editRect.top;
    return ((INT) SendMessage(hEd, EM_CHARFROMPOS, 0, (LPARAM) &pt));
  }
  return -1;
}

INT AnyRichEdit_FirstVisibleLine(HWND hEd)
{
  INT i = AnyRichEdit_FirstVisibleCharIndex(hEd);
  if (i >= 0)
  {
    i = AnyRichEdit_ExLineFromChar(hEd, i);
  }
  return i;
}

TCHAR AnyRichEdit_GetCharAt(HWND hEd, INT nPos)
{
  TCHAR     ch = 0;
  TCHAR     str[4];

  if (AnyRichEdit_GetTextAt(hEd, nPos, 1, str) > 0)
  {
    ch = str[0];
  }

  return ch;
}

BOOL AnyRichEdit_GetEditRect(HWND hEd, RECT* lpRect)
{
  if (!lpRect)
    return FALSE;

  SendMessage(hEd, EM_GETRECT, 0, (LPARAM) lpRect);
  return TRUE;
}

DWORD AnyRichEdit_GetEventMask(HWND hEd)
{
  return ((DWORD) SendMessage(hEd, EM_GETEVENTMASK, 0, 0));
}

INT AnyRichEdit_GetFirstVisibleLine(HWND hEd)
{
  return ((INT) SendMessage(hEd, EM_GETFIRSTVISIBLELINE, 0, 0));
}

INT AnyRichEdit_GetLine(HWND hEd, INT nLine, TCHAR* lpTextBuf, WORD wTextBufSize)
{
  INT   nLen;
  WORD* p;

  p = (WORD*) lpTextBuf;
  *p = wTextBufSize;
  nLen = (INT) SendMessage(hEd, EM_GETLINE, (WPARAM) nLine, (LPARAM) lpTextBuf);
  lpTextBuf[nLen] = 0;
  return nLen;
}

INT AnyRichEdit_GetLineCount(HWND hEd)
{
  return ((INT) SendMessage(hEd, EM_GETLINECOUNT, 0, 0));
}

INT AnyRichEdit_GetTextAt(HWND hEd, INT nPos, INT nLen, TCHAR* lpText)
{
  INT       nRet;
  TEXTRANGE tr;

  tr.lpstrText = lpText;
  tr.chrg.cpMin = nPos;
  tr.chrg.cpMax = nPos + nLen;
  nRet = (INT) SendMessage(hEd, EM_GETTEXTRANGE, 0, (LPARAM) &tr);

  return nRet;
}

INT AnyRichEdit_LastVisibleCharIndex(HWND hEd)
{
  RECT editRect;
  if (AnyRichEdit_GetEditRect(hEd, &editRect))
  {
    POINTL pt;
    pt.x = editRect.right;
    pt.y = editRect.bottom;
    return ((INT) SendMessage(hEd, EM_CHARFROMPOS, 0, (LPARAM) &pt));
  }
  return -1;
}

INT AnyRichEdit_LastVisibleLine(HWND hEd)
{
  INT i = AnyRichEdit_LastVisibleCharIndex(hEd);
  if (i >= 0)
  {
    i = AnyRichEdit_ExLineFromChar(hEd, i);
  }
  return i;
}

INT AnyRichEdit_LineIndex(HWND hEd, INT nLine)
{
  return ((INT) SendMessage(hEd, EM_LINEINDEX, (WPARAM) nLine, 0));
}

INT AnyRichEdit_LineLength(HWND hEd, INT nCharacterIndex)
{
  return ((INT) SendMessage(hEd, EM_LINELENGTH, (WPARAM) nCharacterIndex, 0));
}

void AnyRichEdit_ReplaceSelText(HWND hEd, const TCHAR* cszText, BOOL bCanUndo )
{
  SendMessage(hEd, EM_REPLACESEL, (WPARAM) bCanUndo, (LPARAM) cszText);
}

BOOL AnyRichEdit_SetCharFormat(HWND hEd, DWORD dwMask,
  DWORD dwEffects, COLORREF color, DWORD dwOptions)
{
  BOOL       bRet;
  CHARFORMAT cf;

  cf.cbSize = sizeof(CHARFORMAT);
  cf.dwMask = dwMask;
  cf.dwEffects = dwEffects;
  cf.crTextColor = color;
  bRet = (BOOL) SendMessage(hEd, EM_SETCHARFORMAT, (WPARAM) dwOptions, (LPARAM) &cf);

  return bRet;
}

DWORD AnyRichEdit_SetEventMask(HWND hEd, DWORD nEventMask)
{
  DWORD dwRet;

  dwRet = (DWORD) SendMessage(hEd, EM_SETEVENTMASK, 0, (LPARAM) nEventMask);

  return dwRet;
}

///////////////////////////////////////////////////////////////////////////
// Additional Unicode functions

#ifdef any_ctrl_enable_w_members

  INT AnyRichEdit_ExGetSelPosW(HWND hEd, INT* pnStartPos, INT* pnEndPos)
  {
    CHARRANGE cr;
    SendMessageW(hEd, EM_EXGETSEL, 0, (LPARAM) &cr);
    if (pnStartPos != NULL)
      *pnStartPos = (INT) cr.cpMin;
    if (pnEndPos != NULL)
      *pnEndPos = (INT) cr.cpMax;

    return cr.cpMin;
  }

  void AnyRichEdit_ExLimitTextW(HWND hEd, INT nMaxLength)
  {
    SendMessageW(hEd, EM_EXLIMITTEXT, 0, (LPARAM) nMaxLength);
  }  

  INT AnyRichEdit_ExLineFromCharW(HWND hEd, INT nCharacterIndex)
  {
    return ((INT) SendMessageW(hEd, EM_EXLINEFROMCHAR, 0, (LPARAM) nCharacterIndex));
  }

  void AnyRichEdit_ExSetSelPosW(HWND hEd, INT nStartPos, INT nEndPos)
  {
    CHARRANGE cr;
    cr.cpMin = (LONG) nStartPos;
    cr.cpMax = (LONG) nEndPos;
    SendMessageW(hEd, EM_EXSETSEL, 0, (LPARAM) &cr);
  }

  INT AnyRichEdit_FirstVisibleCharIndexW(HWND hEd)
  {
    RECT editRect;
    if (AnyRichEdit_GetEditRect(hEd, &editRect))
    {
      POINTL pt;
      pt.x = editRect.left;
      pt.y = editRect.top;
      return ((INT) SendMessageW(hEd, EM_CHARFROMPOS, 0, (LPARAM) &pt));
    }
    return -1;
  }

  INT AnyRichEdit_FirstVisibleLineW(HWND hEd)
  {
    INT i = AnyRichEdit_FirstVisibleCharIndexW(hEd);
    if (i >= 0)
    {
      i = AnyRichEdit_ExLineFromCharW(hEd, i);
    }
    return i;
  }
  
  WCHAR AnyRichEdit_GetCharAtW(HWND hEd, INT nPos)
  {
    WCHAR     wch = 0;
    WCHAR     wstr[4];

    if (AnyRichEdit_GetTextAtW(hEd, nPos, 1, wstr) > 0)
    {
      wch = wstr[0];
    }

    return wch;
  }  

  DWORD AnyRichEdit_GetEventMaskW(HWND hEd)
  {
    return ((DWORD) SendMessageW(hEd, EM_GETEVENTMASK, 0, 0));
  }

  INT AnyRichEdit_GetFirstVisibleLineW(HWND hEd)
  {
    return ((INT) SendMessageW(hEd, EM_GETFIRSTVISIBLELINE, 0, 0));
  }

  INT AnyRichEdit_GetLineW(HWND hEd, INT nLine, WCHAR* lpTextBuf, WORD wTextBufSize)
  {
    INT   nLen;
    WORD* p;

    p = (WORD*) lpTextBuf;
    *p = wTextBufSize;
    nLen = (INT) SendMessageW(hEd, EM_GETLINE, (WPARAM) nLine, (LPARAM) lpTextBuf);
    lpTextBuf[nLen] = 0;
    return nLen;
  }
  
  INT AnyRichEdit_GetLineCountW(HWND hEd)
  {
    return ((INT) SendMessageW(hEd, EM_GETLINECOUNT, 0, 0));
  }

  INT AnyRichEdit_GetTextAtW(HWND hEd, INT nPos, INT nLen, WCHAR* lpText)
  {
    INT        nRet;
    TEXTRANGEW trw;

    trw.lpstrText = lpText;
    trw.chrg.cpMin = nPos;
    trw.chrg.cpMax = nPos + nLen;
    nRet = (INT) SendMessageW(hEd, EM_GETTEXTRANGE, 0, (LPARAM) &trw);

    return nRet;
  }

  INT AnyRichEdit_LastVisibleCharIndexW(HWND hEd)
  {
    RECT editRect;
    if (AnyRichEdit_GetEditRect(hEd, &editRect))
    {
      POINTL pt;
      pt.x = editRect.right;
      pt.y = editRect.bottom;
      return ((INT) SendMessageW(hEd, EM_CHARFROMPOS, 0, (LPARAM) &pt));
    }
    return -1;
  }

  INT AnyRichEdit_LastVisibleLineW(HWND hEd)
  {
    INT i = AnyRichEdit_LastVisibleCharIndexW(hEd);
    if (i >= 0)
    {
      i = AnyRichEdit_ExLineFromCharW(hEd, i);
    }
    return i;
  }  
  
  INT AnyRichEdit_LineIndexW(HWND hEd, INT nLine)
  {
    return ((INT) SendMessageW(hEd, EM_LINEINDEX, (WPARAM) nLine, 0));
  }

  INT AnyRichEdit_LineLengthW(HWND hEd, INT nCharacterIndex)
  {
    return ((INT) SendMessageW(hEd, EM_LINELENGTH, (WPARAM) nCharacterIndex, 0));
  }

  void AnyRichEdit_ReplaceSelTextW(HWND hEd, const WCHAR* cszText, BOOL bCanUndo )
  {
    SendMessageW(hEd, EM_REPLACESEL, (WPARAM) bCanUndo, (LPARAM) cszText);
  }

  BOOL AnyRichEdit_SetCharFormatW(HWND hEd, DWORD dwMask,
    DWORD dwEffects, COLORREF color, DWORD dwOptions)
  {
    BOOL       bRet;
    CHARFORMAT cf;

    cf.cbSize = sizeof(CHARFORMAT);
    cf.dwMask = dwMask;
    cf.dwEffects = dwEffects;
    cf.crTextColor = color;
    bRet = (BOOL) SendMessageW(hEd, EM_SETCHARFORMAT, (WPARAM) dwOptions, (LPARAM) &cf);

    return bRet;
  }  

  DWORD AnyRichEdit_SetEventMaskW(HWND hEd, DWORD nEventMask)
  {
    DWORD dwRet;

    dwRet = (DWORD) SendMessageW(hEd, EM_SETEVENTMASK, 0, (LPARAM) nEventMask);

    return dwRet;
  }

#endif

