#include "AutoBrackets.h"
#include "Plugin.h"

//#define _verify_ab_
#undef _verify_ab_

//#define _verify_hl_
#undef _verify_hl_

enum TFileType {
  tftNone = 0,
  tftText,
  tftC_Cpp,
  tftH_Hpp,
  tftPas
};

enum TFileType2 {
  tfmNone           = 0x00,
  tfmComment1       = 0x01,
  tfmHtmlCompatible = 0x02
};

enum TBracketType {
  tbtNone = 0,
  tbtBracket,  //  (
  tbtSquare,   //  [
  tbtBrace,    //  {
  tbtDblQuote, //  "
  tbtSglQuote, //  '
  tbtTag,      //  <
  tbtTag2,

  tbtCount,

  tbtUser
};

const char* strBracketsA[tbtCount - 1] = {
  "()",
  "[]",
  "{}",
  "\"\"",
  "\'\'",
  "<>",
  "</>"
};

const WCHAR* strBracketsW[tbtCount - 1] = {
  L"()",
  L"[]",
  L"{}",
  L"\"\"",
  L"\'\'",
  L"<>",
  L"</>"
};

#define  MAX_EXT             16
#define  MAX_ESCAPED_PREFIX  20
#define  HIGHLIGHT_INDEXES    6
// variables
INT IndexesHighlighted[HIGHLIGHT_INDEXES] = { -1, -1, -1, -1, -1, -1 };
INT IndexesToHighlight[HIGHLIGHT_INDEXES] = { -1, -1, -1, -1, -1, -1 };
INT prevIndexesToHighlight[HIGHLIGHT_INDEXES] = { -1, -1, -1, -1, -1, -1 };
INT CurrentBracketsIndexes[2] = { -1, -1 };

int             nCurrentFileType = tftNone;
int             nCurrentFileType2 = tfmNone;
HWND            hCurrentEditWnd = NULL; // Can be NULL! Use hActualEditWnd.
HWND            hActualEditWnd = NULL; //currentEdit;
BOOL            bBracketsInternalRepaint = FALSE;
INT             nAutoRightBracketPos = -1;
INT             nAutoRightBracketType = tbtNone;

extern HWND     g_hMainWnd;
extern BOOL     g_bOldWindows;
extern BOOL     g_bOldRichEdit; // TRUE means Rich Edit 2.0
extern BOOL     g_bAkelEdit;

extern BOOL     bBracketsHighlightVisibleArea;
extern BOOL     bBracketsRightExistsOK;
extern BOOL     bBracketsDoSingleQuote;
extern BOOL     bBracketsDoTag;
extern BOOL     bBracketsDoTag2;
extern BOOL     bBracketsDoTagIf;
extern BOOL     bBracketsSkipEscaped;
extern BOOL     bBracketsSkipComment1;
extern COLORREF bracketsColourHighlight;
extern char     strHtmlFileExtsA[STR_FILEEXTS_SIZE];
extern wchar_t  strHtmlFileExtsW[STR_FILEEXTS_SIZE];
extern char     strComment1FileExtsA[STR_FILEEXTS_SIZE];
extern wchar_t  strComment1FileExtsW[STR_FILEEXTS_SIZE];

extern DWORD    g_dwOptions[OPT_DWORD_COUNT];

wchar_t         strUserBracketsW[MAX_USER_BRACKETS + 1][4];
char            strUserBracketsA[MAX_USER_BRACKETS + 1][4];
wchar_t         strNextCharOkW[MAX_PREV_NEXT_CHAR_OK_SIZE];
wchar_t         strPrevCharOkW[MAX_PREV_NEXT_CHAR_OK_SIZE];

// functions prototypes
WCHAR char2wchar(const char ch);
void  getEscapedPrefixPos(const INT nOffset, INT* pnPos, INT* pnLen);
BOOL  isEscapedPrefixW(const wchar_t* strW, int len);
void  remove_duplicate_indexes_and_sort(INT* indexes, const INT size /* = HIGHLIGHT_INDEXES */);
BOOL  AutoBracketsFunc(MSGINFO* pmsgi, INT nBracketType, BOOL bOverwriteMode);
BOOL  GetHighlightIndexes(const INT nHighlightIndex, const INT nCharacterPosition);
void  GetPosFromChar(HWND hEd, const INT nCharacterPosition, POINTL* lpPos);
void  HighlightCharacter(const INT nCharacterPosition, const BOOL bHighlight);


/*
int main(void)
{
    // C++ initialization
    currentEdit.m_hWnd = NULL;
    return 0;
}
*/

void* sys_memalloc(int nBytes)
{
    return (void *) GlobalAlloc(GMEM_FIXED, nBytes);
}

void sys_memfree(void* p)
{
    GlobalFree( (HGLOBAL) p );
}

const wchar_t* getBracketsPairW(int nBracketType)
{
    if ( nBracketType < tbtUser )
        return strBracketsW[nBracketType - tbtBracket];

    return strUserBracketsW[nBracketType - tbtUser];
}

const char* getBracketsPairA(int nBracketType)
{
    if ( nBracketType < tbtUser )
        return strBracketsA[nBracketType - tbtBracket];

    return strUserBracketsA[nBracketType - tbtUser];
}

int getLeftBracketType(const wchar_t wch)
{
  int nLeftBracketType = tbtNone;

  switch (wch)
  {
    case L'(' :
      nLeftBracketType = tbtBracket;
      break;
    case L'[' :
      nLeftBracketType = tbtSquare;
      break;
    case L'{' :
      nLeftBracketType = tbtBrace;
      break;
    case L'\"' :
      nLeftBracketType = tbtDblQuote;
      break;
    case L'<' :
      if (bBracketsDoTag)
        nLeftBracketType = tbtTag;
      break;
    case L'\'' :
      if (bBracketsDoSingleQuote)
      {
        nLeftBracketType = tbtSglQuote;
        break;
      } // otherwise go to default (no break here)
    default :
      if (wch != 0)
      {
        int i = 0;
        while (i < MAX_USER_BRACKETS)
        {
          if (strUserBracketsW[i][0] == wch)
          {
            nLeftBracketType = tbtUser + i;
            i = MAX_USER_BRACKETS; // break condition
          }
          else if (strUserBracketsW[i][0] == 0)
          {
            i = MAX_USER_BRACKETS; // break condition
          }
          else
            ++i;
        }
      }
      break;
  }

  return nLeftBracketType;
}

int getRightBracketType(const wchar_t wch)
{
  int nRightBracketType = tbtNone;

  switch (wch)
  {
    case L')' :
      nRightBracketType = tbtBracket;
      break;
    case L']' :
      nRightBracketType = tbtSquare;
      break;
    case L'}' :
      nRightBracketType = tbtBrace;
      break;
    case L'\"' :
      nRightBracketType = tbtDblQuote;
      break;
    case L'>' :
      if (bBracketsDoTag)
        nRightBracketType = tbtTag; 
      // no break here
    case L'/' :
      if (bBracketsDoTag2)
        nRightBracketType = tbtTag2;
      break;
    case L'\'' :
      if (bBracketsDoSingleQuote)
      {
        nRightBracketType = tbtSglQuote;
        break;
      } // otherwise go to default (no break here)
    default :
      if (wch != 0)
      {
        int i = 0;
        while (i < MAX_USER_BRACKETS)
        {
          if (strUserBracketsW[i][1] == wch)
          {
            nRightBracketType = tbtUser + i;
            i = MAX_USER_BRACKETS; // break condition
          }
          else if (strUserBracketsW[i][1] == 0)
          {
            i = MAX_USER_BRACKETS; // break condition
          }
          else
            ++i;
        }
      }
      break;
  }

  return nRightBracketType;
}

BOOL isCharOkW(const wchar_t wch, const wchar_t* pwszChars)
{
  switch (wch)
  {
    case L'\x0D' :
    case L'\x0A' :
    case L'\x00' :
    case L' ' :
    case L'\t' :
      return TRUE;
    default:
      while (*pwszChars)
      {
        if (*pwszChars != wch)
          ++pwszChars;
        else
          return TRUE;
      }
  }
  return FALSE;
}

void OnEditCharPressed(MSGINFO* pmsgi)
{
  WCHAR    wch;
  INT      nBracketType;
  EDITINFO ei;

  SendMessage(g_hMainWnd, AKD_GETEDITINFO, (WPARAM) NULL, (LPARAM) &ei);
  if (ei.bOvertypeMode)
  {
    if (g_dwOptions[OPT_DWORD_OVR_AUTOBR] == 0)
    {
      nAutoRightBracketPos = -1;
      nAutoRightBracketType = tbtNone;
      return;
    }
  }
  
  // verifying if a typed character is a bracket
  if (g_bOldWindows)
  {
    char ch = (char) pmsgi->wParam;
    wch = char2wchar(ch);
  }
  else
  {
    wch = (WCHAR) pmsgi->wParam;
  }
  
  if (pmsgi->hWnd != hActualEditWnd)
  {
    nAutoRightBracketPos = -1;
    nAutoRightBracketType = tbtNone;
  }

  if (nAutoRightBracketPos >= 0)
  {
    // the right bracket has been just added (automatically)
    // but you may duplicate it manually
    INT nEditPos;
    INT nEditEndPos;

    if (g_bOldWindows)
        AnyRichEdit_ExGetSelPos(pmsgi->hWnd, &nEditPos, &nEditEndPos);
    else
        AnyRichEdit_ExGetSelPosW(pmsgi->hWnd, &nEditPos, &nEditEndPos);

    if (nEditPos == nAutoRightBracketPos)
    {
      wchar_t next_wch;

      if (g_bOldWindows)
      {
        char next_ch = (char) AnyRichEdit_GetCharAt(pmsgi->hWnd, nEditPos);
        next_wch = char2wchar(next_ch);
      }
      else
      {
        next_wch = AnyRichEdit_GetCharAtW(pmsgi->hWnd, nEditPos);
      }

      if (getRightBracketType(next_wch) == nAutoRightBracketType)
      {
        if (getRightBracketType(wch) == nAutoRightBracketType)
        {
          // annul pressed character
          pmsgi->wParam = 0;
          ++nEditPos;
          if (nAutoRightBracketType == tbtTag2)
            ++nEditPos;
          if (g_bOldWindows)
            AnyRichEdit_ExSetSelPos(pmsgi->hWnd, nEditPos, nEditPos);
          else
            AnyRichEdit_ExSetSelPosW(pmsgi->hWnd, nEditPos, nEditPos);
          nAutoRightBracketPos = -1;
          nAutoRightBracketType = tbtNone;
          return;
        }

        // one character has been typed, 
        // so the auto-closed right bracket position increased
        ++nAutoRightBracketPos; 
      }
      else
      {
        // not auto-closed right bracket type
        nAutoRightBracketPos = -1;
        nAutoRightBracketType = tbtNone;
      }
    }
    else
    {
      // not auto-closed right bracket position
      nAutoRightBracketPos = -1;
      nAutoRightBracketType = tbtNone;
    }
  }

  nBracketType = getLeftBracketType(wch);
  if (nBracketType != tbtNone)
  {
    // a typed character is a bracket
    if (!AutoBracketsFunc(pmsgi, nBracketType, ei.bOvertypeMode))
    {
      if (nBracketType == tbtTag)
      {
        if (bBracketsDoTag2)
          nBracketType = tbtTag2;
      }
      if (nBracketType == nAutoRightBracketType)
      {
        nAutoRightBracketPos = -1;
        nAutoRightBracketType = tbtNone;
      }
    }
  }
}

void OnEditGetActiveBrackets(MSGINFO* pmsgi, const BOOL bAndHighlight)
{
  INT   i;
  INT   nEditPos;
  INT   nEditEndPos;
  BOOL  bHighlighted;

  if (bBracketsInternalRepaint)
    return;
  
  if (!pmsgi->hWnd)
    return;

  /*
  TCHAR str[256];
  wsprintf(str, "hWnd = %08X, prev = %08X", pmsgi->hWnd, hCurrentEditWnd);
  MessageBox(NULL, str, "OnEditGetActiveBrackets", MB_OK);
  */

  if (hActualEditWnd != pmsgi->hWnd)
  {
    nAutoRightBracketPos = -1;
    nAutoRightBracketType = tbtNone;
  }
  hActualEditWnd = pmsgi->hWnd;
  if (hActualEditWnd != hCurrentEditWnd)
  {
    nCurrentFileType = getFileType(&nCurrentFileType2);
  }

  // getting current position and selection
  if (g_bOldWindows)
    AnyRichEdit_ExGetSelPos(hActualEditWnd, &nEditPos, &nEditEndPos);
  else
    AnyRichEdit_ExGetSelPosW(hActualEditWnd, &nEditPos, &nEditEndPos);

  if (pmsgi->uMsg == WM_LBUTTONUP)
  {
    if (nEditEndPos == nEditPos)
    {
      // it's just simple WM_LBUTTONUP right after WM_LBUTTONDOWN;
      // active brackets have been highlighted already
      return;
    }
    // nEditEndPos != nEditPos,
    // some text has been selected with a mouse
  }

  bHighlighted = FALSE;
  for (i = 0; i < HIGHLIGHT_INDEXES; i++)
  {
    IndexesToHighlight[i] = -1;
  }
  CurrentBracketsIndexes[0] = -1;
  CurrentBracketsIndexes[1] = -1;

  if (GetHighlightIndexes(0, nEditEndPos-1)) // left character
    bHighlighted = TRUE;

  if (g_dwOptions[OPT_DWORD_HLT_BOTHBR] || !bHighlighted)
  {
    if (GetHighlightIndexes(3, nEditEndPos))   // right character
      bHighlighted = TRUE;
  }

  if ((!bHighlighted) && (nEditEndPos != nEditPos))
  {
    if (GetHighlightIndexes(0, nEditPos-1)) // left character
      bHighlighted = TRUE;

    if (g_dwOptions[OPT_DWORD_HLT_BOTHBR] || !bHighlighted)
    {
      if (GetHighlightIndexes(3, nEditPos))   // right character
        bHighlighted = TRUE;
    }
  }
  
  if (bHighlighted)
  {
    for (i = 0; i < HIGHLIGHT_INDEXES; i++)
    {
      prevIndexesToHighlight[i] = IndexesToHighlight[i];
    }
    remove_duplicate_indexes_and_sort(IndexesToHighlight, HIGHLIGHT_INDEXES);
  }
  
  if (bAndHighlight)
  {
    // highlight
    OnEditHighlightActiveBrackets();
  }
  
}

void OnEditHighlightActiveBrackets(void)
{
  int i, j, index;
  int indexesToRemoveHighlight[HIGHLIGHT_INDEXES];

  if (bBracketsInternalRepaint)
    return;

  for (i = 0; i < HIGHLIGHT_INDEXES; i++)
  {
    index = IndexesHighlighted[i];
    for (j = 0; j < HIGHLIGHT_INDEXES; j++)
    {
      if (index == IndexesToHighlight[j])
        break;
    }
    if (j < HIGHLIGHT_INDEXES)
      indexesToRemoveHighlight[i] = -1;
    else
      indexesToRemoveHighlight[i] = index;
  }

  for (i = 0; i < HIGHLIGHT_INDEXES; i++)
  {
    index = indexesToRemoveHighlight[i];
    if (index >= 0)
    {
      IndexesHighlighted[i] = -1; // prevents from WM_PAINT infinite loop
      HighlightCharacter(index, FALSE);
    }
  }

  for (i = 0; i < HIGHLIGHT_INDEXES; i++)
  {
    index = IndexesToHighlight[i];
    IndexesHighlighted[i] = index; // prevents from WM_PAINT infinite loop
    if (index >= 0)
    {
      HighlightCharacter(index, TRUE);
    }
  }

}

BOOL AutoBracketsFunc(MSGINFO* pmsgi, INT nBracketType, BOOL bOverwriteMode)
{
  INT     nEditPos;
  INT     nEditEndPos;
  WCHAR   next_wch;
  BOOL    bPrevCharOK = TRUE;
  BOOL    bNextCharOK = FALSE;

  #ifdef _verify_ab_
    TCHAR str[128];
  #endif

  hActualEditWnd = pmsgi->hWnd;
  if (hActualEditWnd != hCurrentEditWnd)
  {
    nCurrentFileType = getFileType(&nCurrentFileType2);
  }

  if (nBracketType == tbtTag)
  {
    if (bBracketsDoTagIf && !(nCurrentFileType2 & tfmHtmlCompatible))
      return FALSE;
  }

  // getting current position and selection
  if (g_bOldWindows)
    AnyRichEdit_ExGetSelPos(hActualEditWnd, &nEditPos, &nEditEndPos);
  else
    AnyRichEdit_ExGetSelPosW(hActualEditWnd, &nEditPos, &nEditEndPos);

  // is something selected?
  if (nEditEndPos != nEditPos)
  {
    if (g_dwOptions[OPT_DWORD_SEL_AUTOBR] == 0)
    {
      // removing selection
      if (g_bOldWindows)
        AnyRichEdit_ReplaceSelText(hActualEditWnd, "", TRUE);
      else
        AnyRichEdit_ReplaceSelTextW(hActualEditWnd, L"", TRUE);
    }
    else
    {
      if (nBracketType == tbtTag)
      {
        if (bBracketsDoTag2)
          nBracketType = tbtTag2;
      }
        
      if (g_bOldWindows)
      {
        int         nSelLen;
        const char* pBrPairA;
        char*       pTextA;

        nSelLen = nEditEndPos - nEditPos;
        pBrPairA = getBracketsPairA(nBracketType);
        pTextA = (char*)sys_memalloc(sizeof(char)*(lstrlenA(pBrPairA) + nSelLen + 1));
        if (pTextA)
        {
          pTextA[0] = pBrPairA[0];
          nSelLen = AnyRichEdit_GetTextAt(hActualEditWnd, nEditPos, nSelLen, pTextA + 1);
          lstrcpyA(pTextA + nSelLen + 1, pBrPairA + 1);
          AnyRichEdit_ReplaceSelText(hActualEditWnd, pTextA, TRUE);
          sys_memfree(pTextA);
          ++nEditPos;
          ++nEditEndPos;
          AnyRichEdit_ExSetSelPos(hActualEditWnd, nEditPos, nEditEndPos);
        }
      }
      else
      {
        int            nSelLen;
        const wchar_t* pBrPairW;
        wchar_t*       pTextW;
          
        nSelLen = nEditEndPos - nEditPos;
        pBrPairW = getBracketsPairW(nBracketType);
        pTextW = (wchar_t*)sys_memalloc(sizeof(wchar_t)*(lstrlenW(pBrPairW) + nSelLen + 1));
        if (pTextW)
        {
          pTextW[0] = pBrPairW[0];
          nSelLen = AnyRichEdit_GetTextAtW(hActualEditWnd, nEditPos, nSelLen, pTextW + 1);
          lstrcpyW(pTextW + nSelLen + 1, pBrPairW + 1);
          AnyRichEdit_ReplaceSelTextW(hActualEditWnd, pTextW, TRUE);
          sys_memfree(pTextW);
          ++nEditPos;
          ++nEditEndPos;
          AnyRichEdit_ExSetSelPosW(hActualEditWnd, nEditPos, nEditEndPos);
        }
      }

      // annul pressed character
      pmsgi->wParam = 0;
      // no auto bracket position
      nAutoRightBracketPos = -1;
      nAutoRightBracketType = tbtNone;
      return TRUE;
    }
  }

  // getting next character
  if (g_bOldWindows)
  {
    char next_ch = (char) AnyRichEdit_GetCharAt(hActualEditWnd, nEditPos);
    next_wch = char2wchar(next_ch);
  }
  else
  {
    next_wch = AnyRichEdit_GetCharAtW(hActualEditWnd, nEditPos);
  }

  #ifdef _verify_ab_
    wsprintf(str, "next_ch = %04X", next_wch);
    MessageBox(NULL, str, "AutoBracketsFunc", MB_OK);
  #endif

  if ((g_dwOptions[OPT_DWORD_ALL_AUTOBR] & 0x01) ||
      isCharOkW(next_wch, strNextCharOkW))
  {
    int nBrType = getRightBracketType(next_wch);
    if (nBrType == tbtNone)
    {
      bNextCharOK = TRUE;
    }
    else
    {
      if (nBrType == tbtTag2)
      {
        nBrType = tbtTag;
      }
      if (bBracketsRightExistsOK || (nBrType != nBracketType))
      {
        bNextCharOK = TRUE;
      }
    }
    if (bNextCharOK && !bBracketsRightExistsOK)
    {
      if ((next_wch == L'\'') &&
          (getBracketsPairW(nBracketType)[1] == L'\''))
      {
        bNextCharOK = FALSE;
      }
    }
  }

  #ifdef _verify_ab_
    wsprintf(str, "bNextCharOK = %lu", bNextCharOK);
    MessageBox(NULL, str, "AutoBracketsFunc", MB_OK);
  #endif

  if ( bNextCharOK && 
       ((nBracketType == tbtDblQuote) || 
        (nBracketType == tbtSglQuote) ||
        ((nBracketType >= tbtUser) && 
         (getBracketsPairW(nBracketType)[0] == getBracketsPairW(nBracketType)[1]))
       ) 
     )
  {
    wchar_t prev_wch;

    bPrevCharOK = FALSE;
    // getting previous character
    if (g_bOldWindows)
    {
      char prev_ch = (char) AnyRichEdit_GetCharAt(hActualEditWnd, nEditPos-1);
      prev_wch = char2wchar(prev_ch);
    }
    else
    {
      prev_wch = AnyRichEdit_GetCharAtW(hActualEditWnd, nEditPos-1);
    }

    #ifdef _verify_ab_
      wsprintf(str, "prev_ch = %04X", prev_wch);
      MessageBox(NULL, str, "AutoBracketsFunc", MB_OK);
    #endif

    if ((g_dwOptions[OPT_DWORD_ALL_AUTOBR] & 0x02) ||
        isCharOkW(prev_wch, strPrevCharOkW))
    {
      int nBrType = getLeftBracketType(prev_wch);
      if (nBrType == tbtNone)
      {
        bPrevCharOK = TRUE;
      }
      else
      {
        if (nBrType != nBracketType)
        {
          //if (getBracketsPairW(nBrType)[0] != getBracketsPairW(nBrType)[1])
          {
            bPrevCharOK = TRUE;
          }
        }
      }
    }
  }

  if (bPrevCharOK && bNextCharOK && bBracketsSkipEscaped)
  {
    wchar_t szPrefixW[MAX_ESCAPED_PREFIX + 2];
    INT     pos, len;

    getEscapedPrefixPos(nEditPos, &pos, &len);

    if (g_bOldWindows)
    {
      char szPrefixA[MAX_ESCAPED_PREFIX + 2];
    
      len = AnyRichEdit_GetTextAt(hActualEditWnd, pos, len, szPrefixA);
      MultiByteToWideChar(CP_ACP, 0, szPrefixA, -1, szPrefixW, MAX_ESCAPED_PREFIX + 1);
    }
    else
    {
      len = AnyRichEdit_GetTextAtW(hActualEditWnd, pos, len, szPrefixW);
    }

    if (isEscapedPrefixW(szPrefixW, len))
      bPrevCharOK = FALSE;
  }

  if (bPrevCharOK && bNextCharOK)
  {
    // annul pressed character
    pmsgi->wParam = 0;

    if (nBracketType == tbtTag)
    {
      if (bBracketsDoTag2)
        nBracketType = tbtTag2;
    }

    // AkelEdit: the caret position can be out of line length
    if (g_bAkelEdit)
    {
      AECHARINDEX ci;

      ci.lpLine = NULL;
      SendMessage( hActualEditWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ci );
      if (ci.lpLine)
      {
        INT nLineIndex;
        
        if (g_bOldWindows)
          nLineIndex = AnyRichEdit_LineIndex(hActualEditWnd, ci.nLine);
        else
          nLineIndex = AnyRichEdit_LineIndexW(hActualEditWnd, ci.nLine);
        
        nEditPos = nLineIndex + ci.nCharInLine;
      }
    }

    if (bOverwriteMode)
    {
      if (g_dwOptions[OPT_DWORD_OVR_AUTOBR] == 2)
      {
        // overwrite mode: right bracket overwrites existing symbol
        int     nLen;
        int     i;
        wchar_t wsz[8];

        if (g_bOldWindows)
        {
          nLen = lstrlenA(getBracketsPairA(nBracketType));
          wsz[0] = 0;
          AnyRichEdit_GetTextAt(hActualEditWnd, nEditPos, nLen, (char*)wsz);
          for (i = 0; i < nLen; i++)
          {
            switch ( ((char*)wsz)[i] )
            {
              case '\r':
              case '\n':
              case 0:
                nLen = i;
                break;
            }
          }
          AnyRichEdit_ExSetSelPos(hActualEditWnd, nEditPos, nEditPos + nLen);
        }
        else
        {
          nLen = lstrlenW(getBracketsPairW(nBracketType));
          wsz[0] = 0;
          AnyRichEdit_GetTextAtW(hActualEditWnd, nEditPos, nLen, wsz);
          for (i = 0; i < nLen; i++)
          {
            switch ( wsz[i] )
            {
              case L'\r':
              case L'\n':
              case 0:
                nLen = i;
                break;
            }
          }
          AnyRichEdit_ExSetSelPosW(hActualEditWnd, nEditPos, nEditPos + nLen);
        }
      }
    }

    // inserting brackets
    if (g_bOldWindows)
      AnyRichEdit_ReplaceSelText(hActualEditWnd, getBracketsPairA(nBracketType), TRUE);
    else
      AnyRichEdit_ReplaceSelTextW(hActualEditWnd, getBracketsPairW(nBracketType), TRUE);

    // placing cursor between brackets
    nEditPos++;
    if (g_bOldWindows)
      AnyRichEdit_ExSetSelPos(hActualEditWnd, nEditPos, nEditPos);
    else
      AnyRichEdit_ExSetSelPosW(hActualEditWnd, nEditPos, nEditPos);

    nAutoRightBracketPos = nEditPos;
    nAutoRightBracketType = nBracketType;
    return TRUE;
  }

  return FALSE;
}

void updateCurrentBracketsIndexes(INT ind1, INT ind2)
{
  if (ind1 < ind2)
  {
    if ( (CurrentBracketsIndexes[0] < 0) ||
         (ind1 == CurrentBracketsIndexes[0] - 1) ||
         (ind2 == CurrentBracketsIndexes[1] + 1)
       )
    {
      CurrentBracketsIndexes[0] = ind1; // left bracket
      CurrentBracketsIndexes[1] = ind2; // right bracket
    }
  }
  else
  {
    if ( (CurrentBracketsIndexes[0] < 0) ||
         (ind2 == CurrentBracketsIndexes[0] - 1) ||
         (ind1 == CurrentBracketsIndexes[1] + 1)
       )
    {
      CurrentBracketsIndexes[0] = ind2; // left bracket
      CurrentBracketsIndexes[1] = ind1; // right bracket
    }
  }
}

BOOL GetHighlightIndexes(const INT nHighlightIndex, const INT nCharacterPosition)
{
  INT   nBracketType;
  BOOL  bRightBracket;
  WCHAR wch;

  if ( nCharacterPosition < 0 )
    return FALSE;

  /*
  // this does not allow to work with new created file
  if ( nCurrentFileType == tftNone )
    return FALSE;
  */
  
  if ( (prevIndexesToHighlight[0] >= 0) || (prevIndexesToHighlight[3] >= 0) )
  {
    int i;

    for (i = 0; i < HIGHLIGHT_INDEXES; i++)
    {
      if (nCharacterPosition == prevIndexesToHighlight[i])
      {
        int ind1, ind2, k = (i < 3) ? 0 : 3;
        IndexesToHighlight[nHighlightIndex + 0] = prevIndexesToHighlight[k + 0];
        IndexesToHighlight[nHighlightIndex + 1] = prevIndexesToHighlight[k + 1];
        IndexesToHighlight[nHighlightIndex + 2] = prevIndexesToHighlight[k + 2];
        ind1 = IndexesToHighlight[nHighlightIndex + 0];
        if (IndexesToHighlight[nHighlightIndex + 2] < 0)
          ind2 = IndexesToHighlight[nHighlightIndex + 1];
        else
          ind2 = IndexesToHighlight[nHighlightIndex + 2];

        updateCurrentBracketsIndexes(ind1, ind2);
        
        return TRUE;
      }
    }
  }    
  
  if (g_bOldWindows)
  {
    char ch = AnyRichEdit_GetCharAt(hActualEditWnd, nCharacterPosition);
    wch = char2wchar(ch);
  }
  else
  {
    wch = AnyRichEdit_GetCharAtW(hActualEditWnd, nCharacterPosition);
  }

  bRightBracket = FALSE;
  nBracketType = getLeftBracketType(wch);
  if (nBracketType != tbtNone)
  {
    if (getBracketsPairW(nBracketType)[0] == getBracketsPairW(nBracketType)[1])
      nBracketType = tbtNone;
  }
  else
  {
    bRightBracket = TRUE;
    nBracketType = getRightBracketType(wch);
    if (nBracketType != tbtNone)
    {
      if (nBracketType == tbtTag2)
      {
        if (wch == L'>')
          nBracketType = tbtTag;
        else
          nBracketType = tbtNone;
      }
      else
      {
        if (getBracketsPairW(nBracketType)[0] == getBracketsPairW(nBracketType)[1])
          nBracketType = tbtNone;
      }
    }
  }

  if (nBracketType == tbtTag)
  {
    if (bBracketsDoTagIf && !(nCurrentFileType2 & tfmHtmlCompatible))
      return FALSE;
  }

  if (nBracketType != tbtNone && bBracketsSkipEscaped && nCharacterPosition > 0)
  {
    wchar_t szPrefixW[MAX_ESCAPED_PREFIX + 2];
    INT     pos, len;

    getEscapedPrefixPos(nCharacterPosition, &pos, &len);

    if (g_bOldWindows)
    {
      char szPrefixA[MAX_ESCAPED_PREFIX + 2];
      
      len = AnyRichEdit_GetTextAt(hActualEditWnd, pos, len, szPrefixA);
      MultiByteToWideChar(CP_ACP, 0, szPrefixA, -1, szPrefixW, MAX_ESCAPED_PREFIX + 1);
    }
    else
    {
      len = AnyRichEdit_GetTextAtW(hActualEditWnd, pos, len, szPrefixW);
    }

    if (isEscapedPrefixW(szPrefixW, len))
      return FALSE;
  }

  if (nBracketType != tbtNone)
  {
    // we don't want these arrays in the stack ;)
    static TCHAR szLine[0x10000];
    static WCHAR wszLine[0x10000];

    INT   i, iStep, nLen;
    INT   nLine, nMinLine, nMaxLine, nLineStep, nStartLine;
    BOOL  bFound;
    BOOL  bComment;
    WCHAR wchOK, wchFail;
    INT   nFailReferences;
    
    const WCHAR* pcwszLine;
    AECHARINDEX ci;
        
    if (g_bAkelEdit)
      ci.lpLine = NULL;
    else
      pcwszLine = wszLine;

    // the character to search for (another bracket)
    wchOK = getBracketsPairW(nBracketType)[bRightBracket ? 0 : 1];
    wchFail = getBracketsPairW(nBracketType)[bRightBracket ? 1 : 0];
    nFailReferences = 0;

    if (g_bOldWindows)
    {
      nLine = AnyRichEdit_ExLineFromChar(hActualEditWnd, nCharacterPosition);
      nMaxLine = (bBracketsHighlightVisibleArea ?
        AnyRichEdit_LastVisibleLine(hActualEditWnd) : (AnyRichEdit_GetLineCount(hActualEditWnd) - 1));
      nMinLine = (bBracketsHighlightVisibleArea ?
        AnyRichEdit_FirstVisibleLine(hActualEditWnd) : 0);
    }
    else
    {
      nLine = AnyRichEdit_ExLineFromCharW(hActualEditWnd, nCharacterPosition);
      nMaxLine = (bBracketsHighlightVisibleArea ?
        AnyRichEdit_LastVisibleLineW(hActualEditWnd) : (AnyRichEdit_GetLineCountW(hActualEditWnd) - 1));
      nMinLine = (bBracketsHighlightVisibleArea ?
        AnyRichEdit_FirstVisibleLineW(hActualEditWnd) : 0);
    }

    nStartLine = nLine;
    bFound = FALSE;
    bComment = FALSE;
    nLineStep = bRightBracket ? (-1) : 1; // forward or backward
    iStep = bRightBracket ? (-1) : 1;     // forward or backward

    for ( ; bRightBracket ? (nLine >= nMinLine) : (nLine <= nMaxLine); nLine += nLineStep)
    {
      if (g_bAkelEdit)
      {
        if (ci.lpLine)
          ci.lpLine = bRightBracket ? ci.lpLine->prev : ci.lpLine->next;
        else  
          SendMessage( hActualEditWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ci );

        if (ci.lpLine)
        {
          pcwszLine = ci.lpLine->wpLine;
          nLen = ci.lpLine->nLineLen;
        }
        else
          break;
      }
      else
      {
        if (g_bOldWindows)
        {
          nLen = AnyRichEdit_GetLine(hActualEditWnd, nLine, szLine, 0x10000-1);
          // CAnyRichEdit::GetLine sets szLine[nLen] = 0
          MultiByteToWideChar(CP_ACP, 0, szLine, nLen, wszLine, nLen);
          wszLine[nLen] = 0;
        }
        else
        {
          nLen = AnyRichEdit_GetLineW(hActualEditWnd, nLine, wszLine, 0x10000-1);
          // AnyRichEdit_GetLineW sets szLineW[nLen] = 0
        }
      }

      if ( bBracketsSkipComment1 && (nCurrentFileType2 & tfmComment1) )
      {

        // pre-processing current line
        for (i = 0; i < nLen; i++)
        {
          wch = pcwszLine[i];
          if (wch == L'/')
          {
            if ((i+1 < nLen) && (pcwszLine[i+1] == L'/'))
            {
              // this is "//" comment
              if ((i == 0) || (pcwszLine[i-1] != L':')) // not "://"
              {
                //wszLine[i] = 0;
                nLen = i;
                bComment = TRUE;
                break;
              }
            }
          }
        }

      }

      i = bRightBracket ? (nLen - 1) : 0;
      if (nStartLine == nLine)
      {
        INT nLinePosition;

        if (g_bOldWindows)
          nLinePosition = AnyRichEdit_LineIndex(hActualEditWnd, nLine);
        else
          nLinePosition = AnyRichEdit_LineIndexW(hActualEditWnd, nLine);

        nLinePosition = nCharacterPosition - nLinePosition;

        if (bComment) // start line contains a comment...
        {
          if (nLinePosition >= nLen) // ...and current bracket is commented
          {
            //MessageBoxA(NULL,"break","for ( ... nLine ... )",MB_OK);
            break; // break outside the cycle for nLine
          }
        }

        i = bRightBracket ? (nLinePosition - 1) : (nLinePosition + 1);
      }

      for ( ; bRightBracket ? (i >= 0) : (i < nLen); i += iStep)
      {
        wch = pcwszLine[i];
        if (wch == wchOK)
        {
          if (bBracketsSkipEscaped && i > 0)
          {
            INT pos, len;

            getEscapedPrefixPos(i, &pos, &len);
            if (!isEscapedPrefixW(pcwszLine + pos, len))
              --nFailReferences;
          }
          else
            --nFailReferences;
          
          if (nFailReferences < 0)
          {
            bFound = TRUE;
            break;
          }
        }
        else if (wch == wchFail)
        {
          if (bBracketsSkipEscaped && i > 0)
          {
            INT pos, len;

            getEscapedPrefixPos(i, &pos, &len);
            if (!isEscapedPrefixW(pcwszLine + pos, len))
              ++nFailReferences;
          }
          else
            ++nFailReferences;
        }
      }

      if (bFound)
        break;
    }

    if (bFound && (wch == wchOK))
    {
      if (nBracketType == tbtTag)
      {
        WCHAR prev_wch = 0;
        if (bRightBracket)
        {
          if (nCharacterPosition > 0)
          {
            if (g_bOldWindows)
            {
              char ch = AnyRichEdit_GetCharAt(hActualEditWnd, nCharacterPosition-1);
              prev_wch = char2wchar(ch);
            }
            else
            {
              prev_wch = AnyRichEdit_GetCharAtW(hActualEditWnd, nCharacterPosition-1);
            }
          }
        }
        else
        {
          if (i > 0)
          {
            prev_wch = pcwszLine[i-1];
          }
        }
        if (prev_wch == L'/')
        {
          nBracketType = tbtTag2;
        }
      }

      if (g_bOldWindows)
      {
        i += AnyRichEdit_LineIndex(hActualEditWnd, nLine);
      }
      else
      {
        i += AnyRichEdit_LineIndexW(hActualEditWnd, nLine);
      }

      if (nBracketType != tbtTag2)
      {
        IndexesToHighlight[nHighlightIndex] = i;
        IndexesToHighlight[nHighlightIndex+1] = nCharacterPosition;
      }
      else
      {
        IndexesToHighlight[nHighlightIndex] = i;
        IndexesToHighlight[nHighlightIndex+1] =
          bRightBracket ? (nCharacterPosition - 1) : (i - 1);
        IndexesToHighlight[nHighlightIndex+2] = nCharacterPosition;
      }

      updateCurrentBracketsIndexes(i, nCharacterPosition);

      return TRUE;
    }
    else
    {
      //MessageBoxA(NULL,"no indexes","GetHighlightIndexes",MB_OK);
    }

    //return TRUE;
  }

  return FALSE;
}

void GetPosFromChar(HWND hEd, const INT nCharacterPosition, POINTL* lpPos)
{
  if (g_bOldRichEdit)
  {
    // RichEdit 2.0
    LRESULT coord;

    if (g_bOldWindows)
      coord = SendMessageA(hEd, EM_POSFROMCHAR, (WPARAM) nCharacterPosition, 0);
    else
      coord = SendMessageW(hEd, EM_POSFROMCHAR, (WPARAM) nCharacterPosition, 0);
    lpPos->x = LOWORD(coord);
    lpPos->y = HIWORD(coord);
  }
  else
  {
    if (g_bOldWindows)
      SendMessageA(hEd, EM_POSFROMCHAR, (WPARAM) lpPos, (LPARAM) nCharacterPosition);
    else
      SendMessageW(hEd, EM_POSFROMCHAR, (WPARAM) lpPos, (LPARAM) nCharacterPosition);
  }
}

void CopyMemory1(void* dst, const void* src, unsigned int size)
{
  unsigned char* pdst;
  const unsigned char* psrc;
    
  pdst = (unsigned char *) dst;
  psrc = (const unsigned char *) src;
    
  while ( size-- )
  {
    *(pdst++) = *(psrc++);
  }
}

void HighlightCharacter(const INT nCharacterPosition, const BOOL bHighlight)
{
  POINTL  ptBegin, ptEnd;

  if (!hActualEditWnd)
    return;

  if (g_bOldWindows)
  {
    if ((nCharacterPosition < AnyRichEdit_FirstVisibleCharIndex(hActualEditWnd)) ||
        (nCharacterPosition > AnyRichEdit_LastVisibleCharIndex(hActualEditWnd)))
    {
      return;
    }
  }
  else
  {
    if ((nCharacterPosition < AnyRichEdit_FirstVisibleCharIndexW(hActualEditWnd)) ||
        (nCharacterPosition > AnyRichEdit_LastVisibleCharIndexW(hActualEditWnd)))
    {
      return;
    }
  }

  GetPosFromChar(hActualEditWnd, nCharacterPosition, &ptBegin);

  if ((ptBegin.x >= 0) && (ptBegin.y >= 0))
  {
    HDC   hDC;
    RECT  EdRect;

    AnyRichEdit_GetEditRect(hActualEditWnd, &EdRect);

    if (ptBegin.x < EdRect.left)
    {
      // left side of the character can be invisible
      // ptBegin.x = EdRect.left;
      return;
    }

    if (ptBegin.y < EdRect.top)
    {
      // top side of the character can be invisible
      // ptBegin.y = EdRect.top;
      return;
    }

    GetPosFromChar(hActualEditWnd, nCharacterPosition+1, &ptEnd);

    if ((ptEnd.y != ptBegin.y) || (ptEnd.x < 0))
    {
      ptEnd.x = EdRect.right;
    }

    hDC = GetDC(hActualEditWnd);
    if (hDC)
    {
      HRGN        hRgn, hRgnOld;
      HFONT       hFont, hFontOld;
      RECT        rect;
      COLORREF    textColor, bkColor;
      LOGFONTA    lfA;
      LOGFONTW    lfW;
      CHARCOLOR   cc;
      //INT         sel1, sel2;
      char        strA[2];
      wchar_t     strW[2];
      BOOL        bInSelection;

      bBracketsInternalRepaint = TRUE;

      // to repaint a field under the caret
      HideCaret(hActualEditWnd);

      // at first we select a font...
      if (g_bOldWindows)
      {
        LOGFONTA* plfA = (LOGFONTA*) SendMessage(g_hMainWnd, AKD_GETFONT, (WPARAM) hActualEditWnd, 0);
        if (!plfA)
          plfA = (LOGFONTA*) SendMessage(g_hMainWnd, AKD_GETFONT, (WPARAM) NULL, 0);
        CopyMemory1(&lfA, plfA, sizeof(LOGFONTA));
        hFont = CreateFontIndirectA(&lfA);
      }
      else
      {
        LOGFONTW* plfW = (LOGFONTW*) SendMessage(g_hMainWnd, AKD_GETFONT, (WPARAM) hActualEditWnd, 0);
        if (!plfW)
          plfW = (LOGFONTW*) SendMessage(g_hMainWnd, AKD_GETFONT, (WPARAM) NULL, 0);
        CopyMemory1(&lfW, plfW, sizeof(LOGFONTW));
        hFont = CreateFontIndirectW(&lfW);
      }
      hFontOld = (HFONT) SelectObject(hDC, hFont);
      
      // ...then we call GetTextMetrics for this font...
      if (g_bOldWindows)
      {
        TEXTMETRICA tmA;
          
        GetTextMetricsA(hDC, &tmA);
        ptEnd.y = ptBegin.y + tmA.tmHeight;
      }
      else
      {
        TEXTMETRICW tmW;

        GetTextMetricsW(hDC, &tmW);
        ptEnd.y = ptBegin.y + tmW.tmHeight;
      }
      
      // ...finally we can select a region
      ptEnd.x += 3*(ptEnd.x - ptBegin.x); // for some cursive fonts
      if (ptEnd.x > EdRect.right)
      {
        ptEnd.x = EdRect.right;
      }
      if (ptEnd.y > EdRect.bottom)
      {
        ptEnd.y = EdRect.bottom;
      }
      hRgn = CreateRectRgn(ptBegin.x, ptBegin.y, ptEnd.x, ptEnd.y);
      hRgnOld = (HRGN) SelectObject(hDC, hRgn);

      SetBkMode(hDC, TRANSPARENT);
      //bkColor = GetBkColor(hDC);

      //if (g_bOldWindows)
      //  currentEdit.ExGetSelPos(&sel1, &sel2);
      //else
      //  currentEdit.ExGetSelPosW(&sel1, &sel2);

      cc.nCharPos = nCharacterPosition;
      bInSelection = (BOOL) SendMessage(g_hMainWnd, AKD_GETCHARCOLOR, (WPARAM) hActualEditWnd, (LPARAM) &cc);
      bkColor = cc.crBk;
      textColor = cc.crText;
      if (bHighlight)
      {
        //textColor = RGB(16,112,192);
        textColor = bracketsColourHighlight;

        if (bInSelection)
        {
          if (g_dwOptions[OPT_DWORD_HLT_BKGND] != 0)
          {
            BYTE r, g, b;
            
            r = GetRValue(textColor)/2 + GetRValue(bkColor)/2;
            g = GetGValue(textColor)/2 + GetGValue(bkColor)/2;
            b = GetBValue(textColor)/2 + GetBValue(bkColor)/2;

            textColor = RGB(r, g, b);
          }
        }
      }

      SetTextColor(hDC, bkColor);
      rect.left   = ptBegin.x;
      rect.top    = ptBegin.y;
      rect.right  = ptEnd.x;
      rect.bottom = ptEnd.y;
      if (g_bOldWindows)
      {
        if (bHighlight)
        {
          strA[0] = AnyRichEdit_GetCharAt(hActualEditWnd, nCharacterPosition);
          strA[1] = 0;
          DrawTextA(hDC, strA, 1, &rect, 0);

          if (hFontOld)  SelectObject(hDC, hFontOld);
          DeleteObject(hFont);
          lfA.lfWeight = bHighlight ? FW_BOLD : FW_NORMAL;
          hFont = CreateFontIndirectA(&lfA);
          hFontOld = (HFONT) SelectObject(hDC, hFont);
          if (g_dwOptions[OPT_DWORD_HLT_BKGND] == 0)
          {
            SetTextColor(hDC, textColor);
          }
          else
          {
            SetBkMode(hDC, OPAQUE);
            SetBkColor(hDC, textColor);
            SetTextColor(hDC, cc.crText);
          }
          DrawTextA(hDC, strA, 1, &rect, 0);
        }
        else
        {
          InvalidateRgn(hActualEditWnd, hRgn, TRUE);
        }
      }
      else
      {
        if (bHighlight)
        {
          strW[0] = AnyRichEdit_GetCharAtW(hActualEditWnd, nCharacterPosition);
          strW[1] = 0;
          DrawTextW(hDC, strW, 1, &rect, 0);

          if (hFontOld)  SelectObject(hDC, hFontOld);
          DeleteObject(hFont);
          lfW.lfWeight = bHighlight ? FW_BOLD : FW_NORMAL;
          hFont = CreateFontIndirectW(&lfW);
          hFontOld = (HFONT) SelectObject(hDC, hFont);
          if (g_dwOptions[OPT_DWORD_HLT_BKGND] == 0)
          {
            SetTextColor(hDC, textColor);
          }
          else
          {
            SetBkMode(hDC, OPAQUE);
            SetBkColor(hDC, textColor);
            SetTextColor(hDC, cc.crText);
          }
          DrawTextW(hDC, strW, 1, &rect, 0);
        }
        else
        {
          InvalidateRgn(hActualEditWnd, hRgn, TRUE);
        }
      }
      if (hFontOld)  SelectObject(hDC, hFontOld);
      DeleteObject(hFont);
      if (hRgnOld)  SelectObject(hDC, hRgnOld);
      DeleteObject(hRgn);

      ShowCaret(hActualEditWnd);

      ReleaseDC(hActualEditWnd, hDC);

      bBracketsInternalRepaint = FALSE;
    }
  }

}

void RemoveAllHighlightInfo(const BOOL bRepaint)
{
  int i;

  if (bBracketsInternalRepaint)
    return;
  
  for (i = 0; i < HIGHLIGHT_INDEXES; i++)
  {
    IndexesToHighlight[i] = -1; // set no items to highlight
    prevIndexesToHighlight[i] = -1;
  }
  CurrentBracketsIndexes[0] = -1;
  CurrentBracketsIndexes[1] = -1;

  if (bRepaint)
  {
    for (i = 0; i < HIGHLIGHT_INDEXES; i++)
    {
      if (IndexesHighlighted[i] >= 0)  // is any item highlighted?
        break;
    }
    if (i < HIGHLIGHT_INDEXES)
    {
      // removing all highlighting
      OnEditHighlightActiveBrackets();
    } // else there is no highlighted items
  }
  else
  {
    for (i = 0; i < HIGHLIGHT_INDEXES; i++)
    {
      IndexesHighlighted[i] = -1;
    }
  }
}

//---------------------------------------------------------------------------

WCHAR char2wchar(const char ch)
{
  char  str[2] = {0, 0};
  WCHAR wstr[2] = {0, 0};

  str[0] = ch;
  MultiByteToWideChar(CP_ACP, 0, str, 1, wstr, 1);
  return wstr[0];
}

void getEscapedPrefixPos(const INT nOffset, INT* pnPos, INT* pnLen)
{
  if (nOffset > MAX_ESCAPED_PREFIX)
  {
    *pnPos = nOffset - MAX_ESCAPED_PREFIX;
    *pnLen = MAX_ESCAPED_PREFIX;
  }
  else
  {
    *pnPos = 0;
    *pnLen = nOffset;
  }
}

BOOL isEscapedPrefixW(const wchar_t* strW, int len)
{
  int k = 0;
  while ((len > 0) && (strW[--len] == L'\\'))
  {
    ++k;
  }
  return (k % 2) ? TRUE : FALSE;
}

void remove_duplicate_indexes_and_sort(INT* indexes, const INT size)
{
  INT i, j, ind;

  // removing duplicates
  for (i = 0; i < size; i++)
  {
    ind = indexes[i];
    if (ind >= 0)
    {
      for (j = 0; j < size; j++)
      {
        if ((i != j) && (ind == indexes[j]))
        {
          indexes[j] = -1;
        }
      }
    }
  }

  // sorting
  for (i = 0; i < size; i++)
  {
    ind = indexes[i];
    for (j = i+1; j < size; j++)
    {
      if (indexes[j] > ind)
      {
        indexes[i] = indexes[j];
        indexes[j] = ind;  // old value
        ind = indexes[i];  // new value
      }
    }
  }
}

const char* getFileExtA(const char* cszFileNameA)
{
  if (cszFileNameA)
  {
    int pos = lstrlenA(cszFileNameA) - 1;
    if (pos > 0)
    {
      while ((pos >= 0) && (cszFileNameA[pos] != '.'))
      {
        --pos;
      }
      return (cszFileNameA + 1 + pos);
    }
  }
  return NULL;
}

const wchar_t* getFileExtW(const wchar_t* cszFileNameW)
{
  if (cszFileNameW)
  {
    int pos = lstrlenW(cszFileNameW) - 1;
    if (pos > 0)
    {
      while ((pos >= 0) && (cszFileNameW[pos] != L'.'))
      {
        --pos;
      }
      return (cszFileNameW + 1 + pos);
    }
  }
  return NULL;
}

// IMPORTANT!!!  wstr1 and wstr2 must be valid strings!
static int wstr_unsafe_cmp(const wchar_t* wstr1, const wchar_t* wstr2)
{
  while ( (*wstr1) && (*wstr1 == *wstr2) )
  {
    ++wstr1;
    ++wstr2;
  }
  return (*wstr1 - *wstr2);
}

// IMPORTANT!!!  wstr and wsubstr must be valid strings!
static int wstr_unsafe_subcmp(const wchar_t* wstr, const wchar_t* wsubstr)
{
  while ( (*wstr) && (*wstr == *wsubstr) )
  {
    ++wstr;
    ++wsubstr;
  }
  return (*wsubstr) ? (*wstr - *wsubstr) : 0;
}

static BOOL wstr_is_comment1_ext(const wchar_t* szExtW)
{
  if ( szExtW && szExtW[0] )
  {
    int      len, i, n;
    wchar_t  szW[MAX_EXT];

    if (g_bOldWindows)
    {
      len = lstrlenA(strComment1FileExtsA);
      MultiByteToWideChar(CP_ACP, 0, strComment1FileExtsA, len,
        strComment1FileExtsW, STR_FILEEXTS_SIZE - 1);
      strComment1FileExtsW[len] = 0;
    }
    else
    {
      len = lstrlenW(strComment1FileExtsW);
    }

    n = 0;
    i = 0;
    while (i <= len)
    {
      if ( (strComment1FileExtsW[i]) &&
           (strComment1FileExtsW[i] != L';') &&
           (strComment1FileExtsW[i] != L',') &&
           (strComment1FileExtsW[i] != L' ') )
      {
        szW[n++] = strComment1FileExtsW[i];
      }
      else
      {
        if (n > 0)
        {
          szW[n] = 0;
          if (wstr_unsafe_cmp(szExtW, szW) == 0)
            return TRUE;
        }
        n = 0;
      }
      ++i;
    }
  }
  return FALSE;
}

static BOOL wstr_is_html_compatible(const wchar_t* szExtW)
{
  if ( szExtW && szExtW[0] )
  {
    int      len, i, n;
    wchar_t  szW[MAX_EXT];

    if (g_bOldWindows)
    {
      len = lstrlenA(strHtmlFileExtsA);
      MultiByteToWideChar(CP_ACP, 0, strHtmlFileExtsA, len,
        strHtmlFileExtsW, STR_FILEEXTS_SIZE - 1);
      strHtmlFileExtsW[len] = 0;
    }
    else
    {
      len = lstrlenW(strHtmlFileExtsW);
    }

    n = 0;
    i = 0;
    while (i <= len)
    {
      if ( (strHtmlFileExtsW[i]) &&
           (strHtmlFileExtsW[i] != L';') &&
           (strHtmlFileExtsW[i] != L',') &&
           (strHtmlFileExtsW[i] != L' ') )
      {
        szW[n++] = strHtmlFileExtsW[i];
      }
      else
      {
        if (n > 0)
        {
          szW[n] = 0;
          n = 0;
          while (szExtW[n])
          {
            if (wstr_unsafe_subcmp(szExtW + n, szW) == 0)
              return TRUE;
            else
              ++n;
          }
        }
        n = 0;
      }
      ++i;
    }
  }
  return FALSE;
}

int getFileType(int* pnCurrentFileType2)
{
  EDITINFO    ei;
  wchar_t     szExtW[MAX_EXT];
  const void* p = NULL;
  int         nType = tftNone;
  int         nType2 = tfmNone;

  SendMessage(g_hMainWnd, AKD_GETEDITINFO, (WPARAM) hActualEditWnd, (LPARAM) &ei);

  if (g_bOldWindows)
  {
    p = getFileExtA( (const char*) ei.pFile ); // file extension ptr (char *)
    if (p)
    {
      char szExtA[MAX_EXT];
      int  len = 0;
      while ( (len < MAX_EXT - 1) &&
              ((szExtA[len] = ((const char*) p)[len]) != 0) )
      {
        ++len;
      }
      szExtA[len] = 0;
      CharLowerA(szExtA); // MustDie9x does not have CharLowerW !
      MultiByteToWideChar(CP_ACP, 0, szExtA, len, szExtW, MAX_EXT - 1);
      szExtW[len] = 0; // file extension (wchar_t *)
    }
  }
  else
  {
    p = getFileExtW( (const wchar_t*) ei.pFile ); // file extension ptr (wchar_t *)
    if (p)
    {
      int len = 0;
      while ( (len < MAX_EXT - 1) &&
              ((szExtW[len] = ((const wchar_t*) p)[len]) != 0) )
      {
        ++len;
      }
      szExtW[len] = 0; // file extension (wchar_t *)
      CharLowerW(szExtW);
    }
  }

  if (p)
  {
    hCurrentEditWnd = hActualEditWnd;

    //MessageBoxW(NULL, szExtW, L"ext", MB_OK);

    if ( (wstr_unsafe_cmp(szExtW, L"c") == 0)   ||
         (wstr_unsafe_cmp(szExtW, L"cc") == 0) ||
         (wstr_unsafe_cmp(szExtW, L"cpp") == 0) ||
         (wstr_unsafe_cmp(szExtW, L"cxx") == 0) )
    {
      nType = tftC_Cpp;
      nType2 = tfmComment1;
    }
    else if ( (wstr_unsafe_cmp(szExtW, L"h") == 0) ||
              (wstr_unsafe_cmp(szExtW, L"hh") == 0) ||
              (wstr_unsafe_cmp(szExtW, L"hpp") == 0) ||
              (wstr_unsafe_cmp(szExtW, L"hxx") == 0) )
    {
      nType = tftH_Hpp;
      nType2 = tfmComment1;
    }
    else if ( wstr_unsafe_cmp(szExtW, L"pas") == 0 )
    {
      nType = tftPas;
      nType2 = tfmComment1;
    }
    else
    {
      nType = tftText;
      if ( wstr_is_comment1_ext(szExtW) )
      {
        nType2 |= tfmComment1;
      }
      if ( wstr_is_html_compatible(szExtW) )
      {
        nType2 |= tfmHtmlCompatible;
      }
    }

  }
  else
  {
    hCurrentEditWnd = NULL;
  }

  *pnCurrentFileType2 = nType2;
  return nType;
}

void AutoBrackets_Uninitialize(void)
{
  int i;
    
  nCurrentFileType = tftNone;
  nCurrentFileType2 = tfmNone;
  hCurrentEditWnd = NULL;
  hActualEditWnd = NULL;
  nAutoRightBracketPos = -1;
  nAutoRightBracketType = tbtNone;

  RemoveAllHighlightInfo(TRUE);
  
  for (i = 0; i < HIGHLIGHT_INDEXES; i++)
  {
    IndexesHighlighted[i] = -1;
    IndexesToHighlight[i] = -1;
    prevIndexesToHighlight[i] = -1;
  }
  CurrentBracketsIndexes[0] = -1;
  CurrentBracketsIndexes[1] = -1;
}

void copyUniqueCharsOkW(wchar_t* pDst, const wchar_t* pSrc)
{
  int n = 0;

  if (pSrc)
  {
    int     i;
    wchar_t wch;
      
    while ((wch = *pSrc) != 0)
    {
      if ((wch != L' ') && 
          (wch != L'\t') && 
          (wch != L'\x0D') && 
          (wch != L'\x0A'))
      {
        i = 0;
        while ((i < n) && (pDst[i] != wch))
        {
          ++i;
        }
        if (i == n)
        {
          pDst[n] = wch;
          ++n;
        }
      }
      ++pSrc;
    }
  }

  pDst[n] = 0;
}

void setNextCharOkW(const wchar_t* cszNextCharOkW)
{
  copyUniqueCharsOkW(strNextCharOkW, cszNextCharOkW);
}

void setPrevCharOkW(const wchar_t* cszPrevCharOkW)
{
  copyUniqueCharsOkW(strPrevCharOkW, cszPrevCharOkW);
}

void setUserBracketsA(const char* cszUserBracketsA)
{
  strUserBracketsA[0][0] = 0;
  strUserBracketsA[0][1] = 0;
  strUserBracketsA[0][2] = 0;
  strUserBracketsA[0][3] = 0;

  if (cszUserBracketsA && cszUserBracketsA[0])
  {
    int n = 0;
    int i = 0;
    const char* p = cszUserBracketsA;

    while (n < MAX_USER_BRACKETS)
    {
      if ((*p) && (*p != ' ') && (*p != '\t'))
      {
        if (i < 2)
        {
          // brackets pair: 1st or 2nd character
          strUserBracketsA[n][i] = *p;
        }
        else if (i == 2)
        {
          // error: more than 2 characters
          strUserBracketsA[n][0] = 0;
          strUserBracketsA[n][1] = 0;
          strUserBracketsA[n][2] = 0;
          strUserBracketsA[n][3] = 0;
        }
        ++i;
      }
      else
      {
        if (i >= 1)
        {
          if (i == 2)
          {
            // brackets pair completed
            strUserBracketsA[n][2] = 0;
            strUserBracketsA[n][3] = 0;
            ++n;
          }
          // erase incomplete brackets pair
          strUserBracketsA[n][0] = 0;
          strUserBracketsA[n][1] = 0;
          strUserBracketsA[n][2] = 0;
          strUserBracketsA[n][3] = 0;
        }
        i = 0;
        if (*p == 0)
          break;
      }
      ++p;
    }
  }
}

void setUserBracketsW(const wchar_t* cszUserBracketsW)
{
  strUserBracketsW[0][0] = 0;
  strUserBracketsW[0][1] = 0;
  strUserBracketsW[0][2] = 0;
  strUserBracketsW[0][3] = 0;

  if (cszUserBracketsW && cszUserBracketsW[0])
  {
    int n = 0;
    int i = 0;
    const wchar_t* p = cszUserBracketsW;

    while (n < MAX_USER_BRACKETS)
    {
      if ((*p) && (*p != L' ') && (*p != L'\t'))
      {
        if (i < 2)
        {
          // brackets pair: 1st or 2nd character
          strUserBracketsW[n][i] = *p;
        }
        else if (i == 2)
        {
          // error: more than 2 characters
          strUserBracketsW[n][0] = 0;
          strUserBracketsW[n][1] = 0;
          strUserBracketsW[n][2] = 0;
          strUserBracketsW[n][3] = 0;
        }
        ++i;
      }
      else
      {
        if (i >= 1)
        {
          if (i == 2)
          {
            // brackets pair completed
            strUserBracketsW[n][2] = 0;
            strUserBracketsW[n][3] = 0;
            ++n;
          }
          // erase incomplete brackets pair
          strUserBracketsW[n][0] = 0;
          strUserBracketsW[n][1] = 0;
          strUserBracketsW[n][2] = 0;
          strUserBracketsW[n][3] = 0;
        }
        i = 0;
        if (*p == 0)
          break;
      }
      ++p;
    }
  }
}
