#include "base.h"
#include "AutoBrackets.h"
#include "SettingsDlg.h"
#include "resource.h"
#include <richedit.h>
#include "c_base/HexStr.h"
#include "Plugin.h"

//#define _mad_verification_
#undef _mad_verification_

#ifdef _mad_verification_
  TCHAR str[256];
  void ShowMessage(const TCHAR* cszMessage, const TCHAR* cszTitle = _T(""))
  {
    MessageBox(NULL, cszMessage, cszTitle, MB_OK);
  }
#endif

/*
// const
const TCHAR* cszPluginName = _T("XBrackets Plugin");
const TCHAR* cszUnloadPluginQuestion = _T("Plugin is active. Unload it?");
*/

// Functions prototypes
LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void EditParentMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void updatePluginData(PLUGINDATA* pd);
void ReadOptions();
void SaveOptions();

// Global variables
HWND         g_hMainWnd = NULL;
BOOL         g_bOldWindows;
BOOL         g_bOldRichEdit; // TRUE means Rich Edit 2.0
BOOL         g_bMDI;
BOOL         g_bAkelEdit = FALSE;
BOOL         g_bInitialized = FALSE;
WNDPROCDATA* pMainProcData = NULL;
WNDPROCDATA* pEditProcData = NULL;
WNDPROCDATA* pFrameProcData = NULL;
HWND         g_hFocusedEditWnd = NULL;
LANGID       g_LangSystem = 0;

BOOL         bBracketsAutoComplete = TRUE;
BOOL         bBracketsHighlight = TRUE;
BOOL         bBracketsHighlightVisibleArea = TRUE;
BOOL         bBracketsRightExistsOK = FALSE;
BOOL         bBracketsDoSingleQuote = FALSE;
BOOL         bBracketsDoTag = FALSE;
BOOL         bBracketsDoTag2 = FALSE;
BOOL         bBracketsDoTagIf = FALSE;
BOOL         bBracketsSkipEscaped = FALSE;
BOOL         bBracketsSkipComment1 = FALSE;
BOOL         bGoToMatchingBracketTriggered = FALSE;
BOOL         bOpeningNewDocument = FALSE;
COLORREF     bracketsColourHighlight = RGB(0xD0, 0x40, 0x40);
// I also like RGB(0x10, 0x70, 0xC0); // ~ dark blue
// I also like RGB(0xD0, 0x40, 0x40); // ~ red

char         strBracketsColourA[STR_COLOUR_SIZE];
wchar_t      strBracketsColourW[STR_COLOUR_SIZE];

char         strHtmlFileExtsA[STR_FILEEXTS_SIZE] = "htm; xml; php\0";
wchar_t      strHtmlFileExtsW[STR_FILEEXTS_SIZE] = L"htm; xml; php\0";
wchar_t      strHtmlFileExtsW_0[STR_FILEEXTS_SIZE];
char         strComment1FileExtsA[STR_FILEEXTS_SIZE] = "cs; java; js; php\0";
wchar_t      strComment1FileExtsW[STR_FILEEXTS_SIZE] = L"cs; java; js; php\0";
wchar_t      strComment1FileExtsW_0[STR_FILEEXTS_SIZE];
wchar_t      strPluginFuncMainW[STR_PLUGINFUNC_SIZE] = { 0 };

#define      OPT_UNDEFINED_DWORD      0xFEDCBA98
#define      OPTF_AUTOCOMPLETE        0x0001
#define      OPTF_HIGHLIGHT           0x0002
#define      OPTF_VISIBLEAREA         0x0004
#define      OPTF_RIGHTBRACKETOK      0x0008
#define      OPTF_DOSINGLEQUOTE       0x0100
#define      OPTF_DOTAG               0x0200
#define      OPTF_DOTAGIF             0x0400
#define      OPTF_DOTAG2              0x0800
#define      OPTF_SKIPESCAPED         0x1000
#define      OPTF_SKIPCOMMENT1        0x2000

DWORD        opt_dwOptionsFlags0;
DWORD        opt_dwHighlightRGB0;

DWORD        g_dwOptions[OPT_DWORD_COUNT];
DWORD        g_dwOptions0[OPT_DWORD_COUNT];

wchar_t      opt_szUserBracketsW[MAX_USER_BRACKETS*4] = { 0x00AB, 0x00BB, L' ', 0x2018, 0x2019, L' ', 0x201C, 0x201D, 0 };
char         opt_szUserBracketsA[MAX_USER_BRACKETS*4] = { 0xAB, 0xBB, ' ', 0x91, 0x92, ' ', 0x93, 0x94, 0 };
wchar_t      opt_szUserBracketsW_0[MAX_USER_BRACKETS*4];
wchar_t      opt_szNextCharOkW[MAX_PREV_NEXT_CHAR_OK_SIZE] = L".,!?:;<)]}\"\'>/ \x00BB\x2019\x201D\0";
char         opt_szNextCharOkA[MAX_PREV_NEXT_CHAR_OK_SIZE] = ".,!?:;<)]}\"\'>/ \xBB\x92\x94\0";
wchar_t      opt_szNextCharOkW_0[MAX_PREV_NEXT_CHAR_OK_SIZE];
wchar_t      opt_szPrevCharOkW[MAX_PREV_NEXT_CHAR_OK_SIZE] = L"([{<= \x00AB\x2018\x201C\0";
char         opt_szPrevCharOkA[MAX_PREV_NEXT_CHAR_OK_SIZE] = "([{<= \xAB\x91\x93\0";
wchar_t      opt_szPrevCharOkW_0[MAX_PREV_NEXT_CHAR_OK_SIZE];

extern INT     CurrentBracketsIndexes[2];

static const char* cszOptDwordNames[OPT_DWORD_COUNT] = {
  "ovr_autobr",
  "sel_autobr",
  "hlt_bothbr",
  "all_autobr",
  "hlt_bkgnd"
};

static const wchar_t* cwszOptDwordNames[OPT_DWORD_COUNT] = {
  L"ovr_autobr",
  L"sel_autobr",
  L"hlt_bothbr",
  L"all_autobr",
  L"hlt_bkgnd"
};

static LANGID getAkelPadLang(PLUGINDATA* pd)
{
    HANDLE hOptions = (HANDLE) SendMessage( 
      pd->hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM) NULL );

    if ( hOptions ) 
    {
        wchar_t wszModule[MAX_PATH];

        wszModule[0] = 0;
        if ( pd->bOldWindows )
        {
            PLUGINOPTIONA poA;

            poA.pOptionName = "LanguageModule";
            poA.dwType = PO_STRING;
            poA.lpData = (LPBYTE) wszModule;
            poA.dwData = MAX_PATH;
            SendMessage( pd->hMainWnd,
              AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA );
        }
        else
        {
            PLUGINOPTIONW poW;

            poW.pOptionName = L"LanguageModule";
            poW.dwType = PO_STRING;
            poW.lpData = (LPBYTE) wszModule;
            poW.dwData = MAX_PATH * sizeof(wchar_t);
            SendMessage( pd->hMainWnd,
              AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW );
        }
        SendMessage(pd->hMainWnd, AKD_ENDOPTIONS, (WPARAM) hOptions, 0);
        if ( pd->bOldWindows )
        {
            if ( ((LPSTR) wszModule)[0] )
            {
                ((LPSTR) wszModule)[3] = 0;
                CharUpperA( (LPSTR) wszModule );
                if ( lstrcmpA( (LPCSTR) wszModule, "ENG" ) == 0 )
                    return LANG_ENGLISH;
                if ( lstrcmpA( (LPCSTR) wszModule, "RUS" ) == 0 )
                    return LANG_RUSSIAN;
                if ( lstrcmpA( (LPCSTR) wszModule, "UKR" ) == 0 )
                    return LANG_UKRAINIAN;
            }
        }
        else
        {
            if ( wszModule[0] )
            {
                wszModule[3] = 0;
                CharUpperW( wszModule );
                if ( lstrcmpW( wszModule, L"ENG" ) == 0 )
                    return LANG_ENGLISH;
                if ( lstrcmpW( wszModule, L"RUS" ) == 0 )
                    return LANG_RUSSIAN;
                if ( lstrcmpW( wszModule, L"UKR" ) == 0 )
                    return LANG_UKRAINIAN;
            }
        }
    }
    
    return pd->wLangSystem;
}

// Identification
/* extern "C" */
void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion  = AKELDLL;
  pv->dwExeMinVersion3x = MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x = MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName = "XBrackets";
}

// Plugin extern function
/*extern "C"*/
void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *(pd->lpbAutoLoad) = TRUE;
    return;
  }

  updatePluginData(pd);

  if ( pd->bOldWindows )
    lstrcpyA( (LPSTR) strPluginFuncMainW, (LPCSTR) pd->pFunction );
  else
    lstrcpyW( strPluginFuncMainW, (LPCWSTR) pd->pFunction );

  // Is plugin already loaded?
  if (g_bInitialized)
  {
    /*
    if ( MessageBox(g_hMainWnd, cszUnloadPluginQuestion,
           cszPluginName, MB_YESNO | MB_ICONQUESTION) == IDYES )
    */
    {
      Uninitialize(FALSE);
      return;
    }
  }
  else
  {
    Initialize(pd);
  }

  // Stay in memory, and show as active
  pd->nUnload = UD_NONUNLOAD_ACTIVE;
}

/*extern "C"*/
void __declspec(dllexport) Settings(PLUGINDATA *pd)
{
  BOOL     prevInitialized;
  COLORREF prevColourHighlight;
  BOOL     prevBracketsHighlight;
  BOOL     prevBracketsDoTag;
  BOOL     prevBracketsDoTagIf;
  BOOL     prevBracketsDoSingleQuote;
  BOOL     prevBracketsSkipEscaped;
  BOOL     prevBracketsSkipComment1;
  BOOL     bUpdateBracketsHighlight;
  INT_PTR  nRet;
    
  if (pd->lpbAutoLoad)
  {
    *(pd->lpbAutoLoad) = FALSE;
    return;
  }

  updatePluginData(pd);

  #ifdef _mad_verification_
    ShowMessage("Entering Settings...", "void Settings()");
  #endif

  prevInitialized = g_bInitialized;
  if (!g_bInitialized)
  {

    if ( pd->bOldWindows )
    {
      if ( !((LPCSTR) strPluginFuncMainW)[0] )
      {
        int i;

        lstrcpyA( (LPSTR) strPluginFuncMainW, (LPCSTR) pd->pFunction );
        for (i = 0; ((LPCSTR) strPluginFuncMainW)[i] != 0; i++)
        {
          if ( ((LPCSTR) strPluginFuncMainW)[i] == ':' )
          {
            i += 2;
            lstrcpyA( &((LPSTR) strPluginFuncMainW)[i], "Main" );
            break;
          }
        }
      }
    }
    else
    {
      if ( !strPluginFuncMainW[0] )
      {
        int i;
          
        lstrcpyW( strPluginFuncMainW, (LPCWSTR) pd->pFunction );
        for (i = 0; strPluginFuncMainW[i] != 0; i++)
        {
          if (strPluginFuncMainW[i] == L':')
          {
            i += 2;
            lstrcpyW( &strPluginFuncMainW[i], L"Main" );
            break;
          }
        }
      }
    }

    ReadOptions();
  }

  prevColourHighlight = bracketsColourHighlight;
  prevBracketsHighlight = bBracketsHighlight;
  prevBracketsDoTag = bBracketsDoTag;
  prevBracketsDoTagIf = bBracketsDoTagIf;
  prevBracketsDoSingleQuote = bBracketsDoSingleQuote;
  prevBracketsSkipEscaped = bBracketsSkipEscaped;
  prevBracketsSkipComment1 = bBracketsSkipComment1;
  bUpdateBracketsHighlight = FALSE;

  if (g_bOldWindows)
  {
    nRet = DialogBoxA(pd->hInstanceDLL, MAKEINTRESOURCEA(IDD_SETTINGS), 
      pd->hMainWnd, (DLGPROC) SettingsDlgProc);
  }
  else
  {
    nRet = DialogBoxW(pd->hInstanceDLL, MAKEINTRESOURCEW(IDD_SETTINGS), 
      pd->hMainWnd, (DLGPROC) SettingsDlgProc);
  }

  if (nRet && !g_bInitialized)
  {
    SaveOptions();
  }

  if ((prevBracketsDoTag != bBracketsDoTag) ||
      (prevBracketsDoTagIf != bBracketsDoTagIf) ||
      (prevBracketsDoSingleQuote != bBracketsDoSingleQuote) ||
      (prevBracketsSkipEscaped != bBracketsSkipEscaped) ||
      (prevBracketsSkipComment1 != bBracketsSkipComment1))
  {
    bUpdateBracketsHighlight = TRUE;
  }

  if ((prevBracketsHighlight || bGoToMatchingBracketTriggered) && 
      (bUpdateBracketsHighlight || !bBracketsHighlight))
  {
    // brackets were highlighted, now they are not
    RemoveAllHighlightInfo(TRUE);
    bGoToMatchingBracketTriggered = FALSE;
  }

  if (g_bInitialized)
  {
    if (bBracketsHighlight)
    {
      bGoToMatchingBracketTriggered = FALSE;
      if ((!prevBracketsHighlight) || (!prevInitialized) || bUpdateBracketsHighlight)
      {
        MSGINFO msgi = { pd->hWndEdit, WM_PAINT, 0, 0 };

        OnEditGetActiveBrackets(&msgi, TRUE);
      }
      else if (prevColourHighlight != bracketsColourHighlight)
      {
        // brackets highlight colour has been changed
        OnEditHighlightActiveBrackets();
      }
    }
  }

  #ifdef _mad_verification_
    DWORD dwError = GetLastError();
    wsprintf(str,
      _T("DialogBox(0x%08X, ... IDD_SETTINGS ...) returned %ld\n" \
         "GetLastError() returned %lu"),
      pd->hInstanceDLL, nRet, dwError);
    ShowMessage(str, "void Settings()");
  #endif

  // Stay in memory, and show as non-active
  pd->nUnload = UD_NONUNLOAD_NONACTIVE;
}

enum eMatchingBracketAction {
  MBRA_GOTO = 1,
  MBRA_SELTO
};

void DoMatchingBracketAction(const PLUGINDATA *pd, int action)
{
  if (g_bInitialized)
  {
    if ( (!bBracketsHighlight) || 
         (bBracketsHighlightVisibleArea && (CurrentBracketsIndexes[0] < 0)) )
    {
      BOOL    prevBracketsHighlightVisibleArea;
      MSGINFO msgi;
        
      // saving previous value
      prevBracketsHighlightVisibleArea = bBracketsHighlightVisibleArea;
      bBracketsHighlightVisibleArea = FALSE;

      bGoToMatchingBracketTriggered = TRUE;
      msgi.hWnd = pd->hWndEdit;
      msgi.uMsg = WM_PAINT;
      msgi.wParam = 0;
      msgi.lParam = 0;
      OnEditGetActiveBrackets(&msgi, FALSE);

      // restoring previous value
      bBracketsHighlightVisibleArea = prevBracketsHighlightVisibleArea;
    }
      
    if (CurrentBracketsIndexes[0] >= 0)
    {
      CHARRANGE cr = { -1, -1 };
      SendMessage(pd->hWndEdit, EM_EXGETSEL, 0, (LPARAM) &cr);
      if (cr.cpMin == cr.cpMax)
      {
        INT pos = -1;

        if (cr.cpMin == CurrentBracketsIndexes[0] + 1) // after left bracket
          pos = CurrentBracketsIndexes[1];
        else if (cr.cpMin == CurrentBracketsIndexes[1]) // before right bracket
          pos = CurrentBracketsIndexes[0] + 1;
        else if (cr.cpMin == CurrentBracketsIndexes[0]) // before left bracket
          pos = CurrentBracketsIndexes[1] + 1;
        else if (cr.cpMin == CurrentBracketsIndexes[1] + 1) // after right bracket
          pos = CurrentBracketsIndexes[0];

        if (pos >= 0)
        {
          if ( action == MBRA_GOTO )
          {
            cr.cpMin = pos;
            cr.cpMax = pos;
          }
          else if ( action == MBRA_SELTO )
          {
            /*if ( pos < cr.cpMin )
              cr.cpMin = pos;
            else*/
              cr.cpMax = pos;
          }
          SendMessage(pd->hWndEdit, EM_EXSETSEL, 0, (LPARAM) &cr);
        }
      }
    }
  }
}

/*extern "C"*/
void __declspec(dllexport) GoToMatchingBracket(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *(pd->lpbAutoLoad) = FALSE;
    return;
  }

  DoMatchingBracketAction(pd, MBRA_GOTO);

  pd->nUnload = UD_NONUNLOAD_NONACTIVE;
}

/*extern "C"*/
void __declspec(dllexport) SelToMatchingBracket(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *(pd->lpbAutoLoad) = FALSE;
    return;
  }

  DoMatchingBracketAction(pd, MBRA_SELTO);

  pd->nUnload = UD_NONUNLOAD_NONACTIVE;
}

LRESULT CALLBACK NewEditProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  MSGINFO msgi;
  BOOL    bTextChangedLocal = FALSE;

  if ((uMsg == AEM_ADDCLONE) || (uMsg == AEM_DELCLONE))
  {
    LRESULT  lResult = 0;
      
    if (pEditProcData && pEditProcData->NextProc)
      lResult = pEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

    bOpeningNewDocument = FALSE;

    return lResult;
  }
  
  if (uMsg == WM_SETFOCUS)
  {
    EDITINFO ei;
    LRESULT  lResult = 0;
      
    if (pEditProcData && pEditProcData->NextProc)
      lResult = pEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

    ei.hWndEdit = NULL;
    SendMessage(g_hMainWnd, AKD_GETEDITINFO, (WPARAM) NULL, (LPARAM) &ei);

    if (bOpeningNewDocument || (ei.hWndEdit != g_hFocusedEditWnd))
    {
      // Sometimes there is no other way to clear this flag
      bOpeningNewDocument = FALSE;
        
      if (bBracketsHighlight || bGoToMatchingBracketTriggered)
      {
        // RemoveAllHighlightInfo with bRepaint=FALSE fixes strange bug 
        // under MustDie 9x when several MDI windows are repainted
        RemoveAllHighlightInfo(FALSE); 
        bGoToMatchingBracketTriggered = FALSE;
        
        if (bBracketsHighlight)
        {
          if (ei.hWndEdit == hWnd)
          {
            MSGINFO msgi;

            msgi.hWnd = hWnd;
            msgi.uMsg = WM_SETFOCUS;
            msgi.wParam = 0;
            msgi.lParam = 0;

            OnEditGetActiveBrackets(&msgi, TRUE);
          }
        }
      }
      g_hFocusedEditWnd = ei.hWndEdit;
    }
    else
    {
      // Sometimes there is no other way to clear this flag
      bOpeningNewDocument = FALSE;
    }

    return lResult;
  }

  if (!g_bAkelEdit)
  {
    // Rich Edit control
    if (bBracketsHighlight || bGoToMatchingBracketTriggered)
    {
      if ( (uMsg == WM_CHAR && (/*wParam != VK_DELETE &&*/ wParam != VK_BACK)) || 
           (uMsg == WM_KEYDOWN && (wParam == VK_DELETE || wParam == VK_BACK)) ||
           (uMsg == EM_REPLACESEL) ||
           (uMsg == EM_STREAMIN) ||
           (uMsg == EM_PASTESPECIAL) ||
           (uMsg == WM_PASTE) ||
           (uMsg == WM_CUT) ||
           (uMsg == WM_CLEAR) ||
           (uMsg == EM_UNDO) ||
           (uMsg == EM_REDO)
         )
      {
        // before the text is changed: removing the highlighting
        RemoveAllHighlightInfo(TRUE);
        bGoToMatchingBracketTriggered = FALSE;
        bTextChangedLocal = TRUE;
      }
    }
  }
  
  if (bBracketsAutoComplete)
  {
    if (uMsg == WM_CHAR)
    {
      // local operations because only WM_CHAR is processed
      msgi.hWnd   = hWnd;
      msgi.uMsg   = uMsg;
      msgi.wParam = wParam;
      msgi.lParam = lParam;

      OnEditCharPressed(&msgi);

      // local operations because only WM_CHAR is processed
      //uMsg   = msgi.uMsg; // OnEditCharPressed() does not modify uMsg
      wParam = msgi.wParam;
      lParam = msgi.lParam;
    }
  }

  if (bBracketsHighlight)
  {
    if (!g_bAkelEdit)
    {
      // Rich Edit control
      if (bTextChangedLocal)
      {
        LRESULT lResult = 0;

        if (pEditProcData && pEditProcData->NextProc)
          lResult = pEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

        // after the text is changed: updating the highlighting
        msgi.hWnd   = hWnd;
        msgi.uMsg   = uMsg;
        msgi.wParam = wParam;
        msgi.lParam = lParam;

        OnEditGetActiveBrackets(&msgi, TRUE);

        return lResult;
      }
    }
      
    if ((uMsg == WM_PAINT) /* || (uMsg == WM_DISPLAYCHANGE) */ && (hWnd == g_hFocusedEditWnd))
    {
      LRESULT lResult = 0;

      if (pEditProcData && pEditProcData->NextProc)
        lResult = pEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

      OnEditHighlightActiveBrackets();

      return lResult;
    }

  }

  if (pEditProcData && pEditProcData->NextProc)
    return pEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
  else
    return 0;
}

LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  /*
  if (uMsg == AKDN_MAIN_ONSTART_FINISH)
  {
    ReadOptions();
  }
  else
  */
  if (uMsg == AKDN_MAIN_ONFINISH)
  {
    LRESULT lResult = 0;

    if (pMainProcData && pMainProcData->NextProc)
      lResult = pMainProcData->NextProc(hWnd, uMsg, wParam, lParam);

    SaveOptions();
    Uninitialize(TRUE);

    return lResult;
  }

  if (bBracketsHighlight || bGoToMatchingBracketTriggered)
  {
    if ((uMsg == AKDN_OPENDOCUMENT_START) ||
        (uMsg == AKDN_EDIT_ONSTART) ||
        (uMsg == WM_COMMAND && LOWORD(wParam) == IDM_FILE_NEW))
    {
      // AKDN_EDIT_ONSTART - required to set bOpeningNewDocument before EN_SELCHANGE
      // WM_COMMAND, IDM_FILE_NEW - required for Sessions plugin
      bOpeningNewDocument = TRUE;
    }
    else if (uMsg == AKDN_OPENDOCUMENT_FINISH)
    {
      LRESULT lResult = 0;

      if (pMainProcData && pMainProcData->NextProc)
        lResult = pMainProcData->NextProc(hWnd, uMsg, wParam, lParam);

      bOpeningNewDocument = FALSE;
      if (bBracketsHighlight)
      {
        MSGINFO msgi = { (HWND) wParam, WM_PAINT, 0, 0 };
        OnEditGetActiveBrackets(&msgi, TRUE);
      }

      return lResult;
    }
    else
      EditParentMessages(hWnd, uMsg, wParam, lParam);
  }

  if (pMainProcData && pMainProcData->NextProc)
    return pMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
  else
    return 0;
}

LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (bBracketsHighlight || bGoToMatchingBracketTriggered)
  {
    /*
    if (uMsg == WM_MDIACTIVATE)
    {
      LRESULT lResult = 0;

      if (pFrameProcData && pFrameProcData->NextProc)
        lResult = pFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);

      OnActiveDocumentChanged( GetFocus() );

      return lResult;
    }
    */
  
    EditParentMessages(hWnd, uMsg, wParam, lParam);
  }

  if (pFrameProcData && pFrameProcData->NextProc)
    return pFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);
  else
    return 0;
}

void EditParentMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_NOTIFY)
  {
    if (wParam == ID_EDIT)
    {
      if (((NMHDR *)lParam)->code == EN_SELCHANGE)
      {
        if (g_hFocusedEditWnd == ((NMHDR *)lParam)->hwndFrom)
        {
          MSGINFO msgi;

          msgi.hWnd = ((NMHDR *)lParam)->hwndFrom;
          msgi.uMsg = WM_PAINT;
          msgi.wParam = 0;
          msgi.lParam = 0;
          
          if (bBracketsHighlight && !bOpeningNewDocument)
          {
            OnEditGetActiveBrackets(&msgi, TRUE);
          }
        }
      }
      else if (g_bAkelEdit)
      {
        // AkelEdit control
        if (((NMHDR *)lParam)->code == AEN_TEXTCHANGING)
        {
          // before the text is changed: removing the highlighting
          RemoveAllHighlightInfo(TRUE);
          bGoToMatchingBracketTriggered = FALSE;
        }
        else if (((NMHDR *)lParam)->code == AEN_TEXTCHANGED)
        {
          // after the text is changed: updating the highlighting
          if (bBracketsHighlight)
          {
            MSGINFO msgi;

            msgi.hWnd = ((NMHDR *)lParam)->hwndFrom;
            msgi.uMsg = WM_PAINT;
            msgi.wParam = 0;
            msgi.lParam = 0;

            OnEditGetActiveBrackets(&msgi, TRUE);
          }
        }
      }
    }
  }
  //return 0;
}

void updatePluginData(PLUGINDATA* pd)
{
  g_hMainWnd = pd->hMainWnd;
  g_bMDI = pd->bMDI;
  g_bOldWindows = pd->bOldWindows;
  g_bOldRichEdit = pd->bOldRichEdit;
  g_bAkelEdit = pd->bAkelEdit;
  g_LangSystem = getAkelPadLang(pd);
}

void Initialize(PLUGINDATA* pd)
{
  #ifdef _mad_verification_
    ShowMessage("Entering Initialize...", "void Initialize()");
  #endif

  #ifdef _mad_verification_
    wsprintf(str,
      _T("hMainWnd = 0x%08X\n" \
         "bMDI = %lu\n" \
         "bOldWindows = %lu\n" \
         "bOldRichEdit = %lu"),
      g_hMainWnd, g_bMDI, g_bOldWindows, g_bOldRichEdit);
    ShowMessage(str, "void Initialize()");
  #endif

  if (!g_bInitialized)
  {
    ReadOptions();

    // SubClass main window
    pMainProcData = NULL;
    SendMessage( g_hMainWnd, AKD_SETMAINPROC,
      (WPARAM) NewMainProc, (LPARAM) &pMainProcData );

    // SubClass edit window
    pEditProcData = NULL;
    SendMessage( g_hMainWnd, AKD_SETEDITPROC,
      (WPARAM) NewEditProc, (LPARAM) &pEditProcData );

    if (g_bMDI)
    {
      // SubClass frame window
      pFrameProcData = NULL;
      SendMessage( g_hMainWnd, AKD_SETFRAMEPROC,
        (WPARAM) NewFrameProc, (LPARAM) &pFrameProcData );
    }

    g_hFocusedEditWnd = pd->hWndEdit;

    if (bBracketsHighlight)
    {
      MSGINFO msgi = { pd->hWndEdit, WM_PAINT, 0, 0 };

      OnEditGetActiveBrackets(&msgi, TRUE);
    }

    #ifdef _mad_verification_
      wsprintf(str,
        _T("bInitialized = FALSE\n" \
           "pMainProcData = 0x%08X\n" \
           "pEditProcData = 0x%08X"),
        pMainProcData, pEditProcData);
      ShowMessage(str, "void Initialize()");
      } else {
      ShowMessage("bInitialized = TRUE", "void Initialize()");
    #endif
  }

  g_bInitialized = TRUE;
}

void Uninitialize(BOOL bIsExiting )
{
  if (g_bInitialized)
  {
    g_bInitialized = FALSE;

    g_hFocusedEditWnd = NULL;

    if (!bIsExiting)
    {
      if (bBracketsHighlight || bGoToMatchingBracketTriggered)
      {
        RemoveAllHighlightInfo(TRUE);
        bGoToMatchingBracketTriggered = FALSE;
      }
    }

    if (pMainProcData)
    {
      // Remove subclass (main window)
      SendMessage( g_hMainWnd, AKD_SETMAINPROC,
        (WPARAM) NULL, (LPARAM) &pMainProcData );
      pMainProcData = NULL;
    }

    if (pEditProcData)
    {
      // Remove subclass (edit window)
      SendMessage( g_hMainWnd, AKD_SETEDITPROC,
        (WPARAM) NULL, (LPARAM) &pEditProcData );
      pEditProcData = NULL;
    }

    if (g_bMDI && pFrameProcData)
    {
      // Remove subclass (frame window)
      SendMessage( g_hMainWnd, AKD_SETFRAMEPROC,
        (WPARAM) NULL, (LPARAM) &pFrameProcData );
      pFrameProcData = NULL;
    }

    if (!bIsExiting)
    {
      SaveOptions();

      AutoBrackets_Uninitialize();
    }
  }
}

void ReadOptions()
{
  DWORD dwValue;
  int   i;

  strHtmlFileExtsW_0[0] = 0;
  strComment1FileExtsW_0[0] = 0;
  opt_szNextCharOkW_0[0] = 0;
  opt_szPrevCharOkW_0[0] = 0;
  opt_szUserBracketsW_0[0] = 0;

  for (i = 0; i < OPT_DWORD_COUNT; i++)
  {
    g_dwOptions[i] = OPT_UNDEFINED_DWORD;
  }

  if (g_bOldWindows)
  {
    HANDLE        hOptions;
    PLUGINOPTIONA poA;

    hOptions = (HANDLE) SendMessage(g_hMainWnd,
      AKD_BEGINOPTIONS, POB_READ, (LPARAM) "XBrackets");

    if (hOptions)
    {
      poA.pOptionName = "OptionsFlags";
      poA.lpData = (BYTE*) &dwValue;
      poA.dwData = sizeof(DWORD);
      poA.dwType = PO_DWORD;
      dwValue = OPT_UNDEFINED_DWORD;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
      opt_dwOptionsFlags0 = dwValue;  // opt_dwOptionsFlags0 value is set

      poA.pOptionName = "HighlightRGB";
      poA.lpData = (BYTE*) &dwValue;
      poA.dwData = sizeof(DWORD);
      poA.dwType = PO_BINARY;
      dwValue = OPT_UNDEFINED_DWORD;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
      opt_dwHighlightRGB0 = dwValue;
      if (opt_dwHighlightRGB0 != OPT_UNDEFINED_DWORD)
      {
        /*c_base::*/byte_t btColour[4];

        btColour[0] = GetRValue( (COLORREF) opt_dwHighlightRGB0 );
        btColour[1] = GetGValue( (COLORREF) opt_dwHighlightRGB0 );
        btColour[2] = GetBValue( (COLORREF) opt_dwHighlightRGB0 );
        bracketsColourHighlight = RGB(btColour[0], btColour[1], btColour[2]);
        /*c_base::*/buf2hexstr(btColour, 3, strBracketsColourA, STR_COLOUR_SIZE, " ");
      }

      poA.pOptionName = "HtmlFileExts";
      poA.lpData = (BYTE*) strHtmlFileExtsA;
      poA.dwData = (STR_FILEEXTS_SIZE - 1)*sizeof(char);
      poA.dwType = PO_STRING;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
      CharLowerA(strHtmlFileExtsA);
      lstrcpyA( (LPSTR) strHtmlFileExtsW_0, strHtmlFileExtsA );

      poA.pOptionName = "Comment1FileExts";
      poA.lpData = (BYTE*) strComment1FileExtsA;
      poA.dwData = (STR_FILEEXTS_SIZE - 1)*sizeof(char);
      poA.dwType = PO_STRING;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
      CharLowerA(strComment1FileExtsA);
      lstrcpyA( (LPSTR) strComment1FileExtsW_0, strComment1FileExtsA );

      poA.pOptionName = "next_char_ok";
      poA.lpData = (BYTE*) opt_szNextCharOkW_0;
      poA.dwData = (MAX_PREV_NEXT_CHAR_OK_SIZE - 1)*sizeof(char);
      poA.dwType = PO_STRING;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);

      poA.pOptionName = "prev_char_ok";
      poA.lpData = (BYTE*) opt_szPrevCharOkW_0;
      poA.dwData = (MAX_PREV_NEXT_CHAR_OK_SIZE - 1)*sizeof(char);
      poA.dwType = PO_STRING;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
      
      poA.pOptionName = "user_brpairs";
      poA.lpData = (BYTE*) opt_szUserBracketsW_0;
      poA.dwData = (MAX_USER_BRACKETS*4 - 1)*sizeof(char);
      poA.dwType = PO_STRING;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);

      for (i = 0; i < OPT_DWORD_COUNT; i++)
      {
        poA.pOptionName = (char*) cszOptDwordNames[i];
        poA.lpData = (BYTE*) &g_dwOptions[i];
        poA.dwData = sizeof(DWORD);
        poA.dwType = PO_DWORD;
        g_dwOptions[i] = OPT_UNDEFINED_DWORD;
        SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
        g_dwOptions0[i] = g_dwOptions[i];
      }

      SendMessage(g_hMainWnd, AKD_ENDOPTIONS, (WPARAM) hOptions, 0);
    }

  }
  else
  {
    HANDLE        hOptions;
    PLUGINOPTIONW poW;

    hOptions = (HANDLE) SendMessage(g_hMainWnd,
      AKD_BEGINOPTIONS, POB_READ, (LPARAM) L"XBrackets");

    if (hOptions)
    {
      poW.pOptionName = L"OptionsFlags";
      poW.lpData = (BYTE*) &dwValue;
      poW.dwData = sizeof(DWORD);
      poW.dwType = PO_DWORD;
      dwValue = OPT_UNDEFINED_DWORD;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
      opt_dwOptionsFlags0 = dwValue;  // opt_dwOptionsFlags0 value is set

      poW.pOptionName = L"HighlightRGB";
      poW.lpData = (BYTE*) &dwValue;
      poW.dwData = sizeof(DWORD);
      poW.dwType = PO_BINARY;
      dwValue = OPT_UNDEFINED_DWORD;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
      opt_dwHighlightRGB0 = dwValue;
      if (opt_dwHighlightRGB0 != OPT_UNDEFINED_DWORD)
      {
        /*c_base::*/byte_t btColour[4];

        btColour[0] = GetRValue( (COLORREF) opt_dwHighlightRGB0 );
        btColour[1] = GetGValue( (COLORREF) opt_dwHighlightRGB0 );
        btColour[2] = GetBValue( (COLORREF) opt_dwHighlightRGB0 );
        bracketsColourHighlight = RGB(btColour[0], btColour[1], btColour[2]);
        /*c_base::*/buf2hexstrw(btColour, 3, strBracketsColourW, STR_COLOUR_SIZE, L" ");
      }

      poW.pOptionName = L"HtmlFileExts";
      poW.lpData = (BYTE*) strHtmlFileExtsW;
      poW.dwData = (STR_FILEEXTS_SIZE - 1)*sizeof(wchar_t);
      poW.dwType = PO_STRING;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
      CharLowerW(strHtmlFileExtsW);
      lstrcpyW(strHtmlFileExtsW_0, strHtmlFileExtsW);

      poW.pOptionName = L"Comment1FileExts";
      poW.lpData = (BYTE*) strComment1FileExtsW;
      poW.dwData = (STR_FILEEXTS_SIZE - 1)*sizeof(wchar_t);
      poW.dwType = PO_STRING;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
      CharLowerW(strComment1FileExtsW);
      lstrcpyW(strComment1FileExtsW_0, strComment1FileExtsW);

      poW.pOptionName = L"next_char_ok";
      poW.lpData = (BYTE*) opt_szNextCharOkW_0;
      poW.dwData = (MAX_PREV_NEXT_CHAR_OK_SIZE - 1)*sizeof(wchar_t);
      poW.dwType = PO_STRING;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);

      poW.pOptionName = L"prev_char_ok";
      poW.lpData = (BYTE*) opt_szPrevCharOkW_0;
      poW.dwData = (MAX_PREV_NEXT_CHAR_OK_SIZE - 1)*sizeof(wchar_t);
      poW.dwType = PO_STRING;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);

      poW.pOptionName = L"user_brpairs";
      poW.lpData = (BYTE*) opt_szUserBracketsW_0;
      poW.dwData = (MAX_USER_BRACKETS*4 - 1)*sizeof(wchar_t);
      poW.dwType = PO_STRING;
      SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);

      for (i = 0; i < OPT_DWORD_COUNT; i++)
      {
        poW.pOptionName = (wchar_t*) cwszOptDwordNames[i];
        poW.lpData = (BYTE*) &g_dwOptions[i];
        poW.dwData = sizeof(DWORD);
        poW.dwType = PO_DWORD;
        g_dwOptions[i] = OPT_UNDEFINED_DWORD;
        SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
        g_dwOptions0[i] = g_dwOptions[i];
      }

      SendMessage(g_hMainWnd, AKD_ENDOPTIONS, (WPARAM) hOptions, 0);
    }

  }

  // opt_dwOptionsFlags0 value has been set before
  if (opt_dwOptionsFlags0 != OPT_UNDEFINED_DWORD)
  {
    bBracketsAutoComplete =
      ((opt_dwOptionsFlags0 & OPTF_AUTOCOMPLETE) == OPTF_AUTOCOMPLETE);
    bBracketsHighlight =
      ((opt_dwOptionsFlags0 & OPTF_HIGHLIGHT) == OPTF_HIGHLIGHT);
    bBracketsHighlightVisibleArea =
      ((opt_dwOptionsFlags0 & OPTF_VISIBLEAREA) == OPTF_VISIBLEAREA);
    bBracketsRightExistsOK =
      ((opt_dwOptionsFlags0 & OPTF_RIGHTBRACKETOK) == OPTF_RIGHTBRACKETOK);
    bBracketsDoSingleQuote =
      ((opt_dwOptionsFlags0 & OPTF_DOSINGLEQUOTE) == OPTF_DOSINGLEQUOTE);
    bBracketsDoTag =
      ((opt_dwOptionsFlags0 & OPTF_DOTAG) == OPTF_DOTAG);
    bBracketsDoTag2 =
      ((opt_dwOptionsFlags0 & OPTF_DOTAG2) == OPTF_DOTAG2);
    bBracketsDoTagIf =
      ((opt_dwOptionsFlags0 & OPTF_DOTAGIF) == OPTF_DOTAGIF);
    bBracketsSkipEscaped =
      ((opt_dwOptionsFlags0 & OPTF_SKIPESCAPED) == OPTF_SKIPESCAPED);
    bBracketsSkipComment1 = 
      ((opt_dwOptionsFlags0 & OPTF_SKIPCOMMENT1) == OPTF_SKIPCOMMENT1);
  }

  if (opt_szNextCharOkW_0[0] != 0)
  {
    if (g_bOldWindows)
    {
      lstrcpyA(opt_szNextCharOkA, (LPCSTR)opt_szNextCharOkW_0);
      MultiByteToWideChar(CP_ACP, 0, opt_szNextCharOkA, -1, 
        opt_szNextCharOkW, MAX_PREV_NEXT_CHAR_OK_SIZE - 1);
    }
    else
    {
      lstrcpyW(opt_szNextCharOkW, opt_szNextCharOkW_0);
    }
  }
  setNextCharOkW(opt_szNextCharOkW);

  if (opt_szPrevCharOkW_0[0] != 0)
  {
    if (g_bOldWindows)
    {
      lstrcpyA(opt_szPrevCharOkA, (LPCSTR)opt_szPrevCharOkW_0);
      MultiByteToWideChar(CP_ACP, 0, opt_szPrevCharOkA, -1, 
        opt_szPrevCharOkW, MAX_PREV_NEXT_CHAR_OK_SIZE - 1);
    }
    else
    {
      lstrcpyW(opt_szPrevCharOkW, opt_szPrevCharOkW_0);
    }
  }
  setPrevCharOkW(opt_szPrevCharOkW);
  
  if (opt_szUserBracketsW_0[0] != 0)
  {
    if (g_bOldWindows)
    {
      lstrcpyA(opt_szUserBracketsA, (LPCSTR)opt_szUserBracketsW_0);
      MultiByteToWideChar(CP_ACP, 0, opt_szUserBracketsA, -1, 
        opt_szUserBracketsW, MAX_USER_BRACKETS*4 - 1);
    }
    else
    {
      lstrcpyW(opt_szUserBracketsW, opt_szUserBracketsW_0);
    }
  }

  if (g_bOldWindows)
  {
    setUserBracketsA(opt_szUserBracketsA);
    setUserBracketsW(opt_szUserBracketsW);
  }
  else
  {
    setUserBracketsW(opt_szUserBracketsW);
  }

  if (g_dwOptions[OPT_DWORD_OVR_AUTOBR] == OPT_UNDEFINED_DWORD)
    g_dwOptions[OPT_DWORD_OVR_AUTOBR] = 0;
  if (g_dwOptions[OPT_DWORD_SEL_AUTOBR] == OPT_UNDEFINED_DWORD)
    g_dwOptions[OPT_DWORD_SEL_AUTOBR] = 0;
  if (g_dwOptions[OPT_DWORD_HLT_BOTHBR] == OPT_UNDEFINED_DWORD)
    g_dwOptions[OPT_DWORD_HLT_BOTHBR] = 1;
  if (g_dwOptions[OPT_DWORD_ALL_AUTOBR] == OPT_UNDEFINED_DWORD)
    g_dwOptions[OPT_DWORD_ALL_AUTOBR] = 0;
  if (g_dwOptions[OPT_DWORD_HLT_BKGND] == OPT_UNDEFINED_DWORD)
    g_dwOptions[OPT_DWORD_HLT_BKGND] = 0;

}

void SaveOptions()
{
  DWORD dwNewOptionsFlags;
  DWORD dwNewHighlightRGB;
  int   nStrCmpHtmlExts;
  int   nStrCmpComment1Exts;
  int   nStrCmpNextCharOk;
  int   nStrCmpPrevCharOk;
  int   nStrCmpUsrBrPairs;
  int   i;

  dwNewHighlightRGB = (DWORD) bracketsColourHighlight;

  dwNewOptionsFlags = 0;
  if (bBracketsAutoComplete)
    dwNewOptionsFlags |= OPTF_AUTOCOMPLETE;
  if (bBracketsHighlight)
    dwNewOptionsFlags |= OPTF_HIGHLIGHT;
  if (bBracketsHighlightVisibleArea)
    dwNewOptionsFlags |= OPTF_VISIBLEAREA;
  if (bBracketsRightExistsOK)
    dwNewOptionsFlags |= OPTF_RIGHTBRACKETOK;
  if (bBracketsDoSingleQuote)
    dwNewOptionsFlags |= OPTF_DOSINGLEQUOTE;
  if (bBracketsDoTag)
    dwNewOptionsFlags |= OPTF_DOTAG;
  if (bBracketsDoTag2)
    dwNewOptionsFlags |= OPTF_DOTAG2;
  if (bBracketsDoTagIf)
    dwNewOptionsFlags |= OPTF_DOTAGIF;
  if (bBracketsSkipEscaped)
    dwNewOptionsFlags |= OPTF_SKIPESCAPED;
  if (bBracketsSkipComment1) 
    dwNewOptionsFlags |= OPTF_SKIPCOMMENT1;

  if (g_bOldWindows)
  {
    nStrCmpHtmlExts = lstrcmpiA( (LPCSTR) strHtmlFileExtsW_0, strHtmlFileExtsA );
    nStrCmpComment1Exts = lstrcmpiA( (LPCSTR) strComment1FileExtsW_0, strComment1FileExtsA );
    nStrCmpNextCharOk = lstrcmpA( (LPCSTR) opt_szNextCharOkW_0, opt_szNextCharOkA );
    nStrCmpPrevCharOk = lstrcmpA( (LPCSTR) opt_szPrevCharOkW_0, opt_szPrevCharOkA );
    nStrCmpUsrBrPairs = lstrcmpA( (LPCSTR) opt_szUserBracketsW_0, opt_szUserBracketsA );
  }
  else
  {
    nStrCmpHtmlExts = lstrcmpiW( strHtmlFileExtsW_0, strHtmlFileExtsW );
    nStrCmpComment1Exts = lstrcmpiW( strComment1FileExtsW_0, strComment1FileExtsW );
    nStrCmpNextCharOk = lstrcmpW( opt_szNextCharOkW_0, opt_szNextCharOkW );
    nStrCmpPrevCharOk = lstrcmpW( opt_szPrevCharOkW_0, opt_szPrevCharOkW );
    nStrCmpUsrBrPairs = lstrcmpW( opt_szUserBracketsW_0, opt_szUserBracketsW );
  }

  for (i = 0; i < OPT_DWORD_COUNT; i++)
  {
    if (g_dwOptions[i] != g_dwOptions0[i])
      break;
  }

  if ((i < OPT_DWORD_COUNT) ||
      (dwNewOptionsFlags != opt_dwOptionsFlags0) ||
      (dwNewHighlightRGB != opt_dwHighlightRGB0) ||
      (nStrCmpHtmlExts != 0) ||
      (nStrCmpComment1Exts != 0) ||
      (nStrCmpNextCharOk != 0) ||
      (nStrCmpPrevCharOk != 0) ||
      (nStrCmpUsrBrPairs != 0))
  {
    if (g_bOldWindows)
    {
      HANDLE        hOptions;
      PLUGINOPTIONA poA;

      hOptions = (HANDLE) SendMessage(g_hMainWnd,
        AKD_BEGINOPTIONS, POB_SAVE, (LPARAM) "XBrackets");

      if (hOptions)
      {
        if (dwNewOptionsFlags != opt_dwOptionsFlags0)
        {
          poA.pOptionName = "OptionsFlags";
          poA.lpData = (BYTE*) &dwNewOptionsFlags;
          poA.dwData = sizeof(DWORD);
          poA.dwType = PO_DWORD;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
          opt_dwOptionsFlags0 = dwNewOptionsFlags;
        }
        if (dwNewHighlightRGB != opt_dwHighlightRGB0)
        {
          poA.pOptionName = "HighlightRGB";
          poA.lpData = (BYTE*) &dwNewHighlightRGB;
          poA.dwData = sizeof(DWORD);
          poA.dwType = PO_BINARY;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
          opt_dwHighlightRGB0 = dwNewHighlightRGB;
        }
        if (nStrCmpHtmlExts != 0)
        {
          poA.pOptionName = "HtmlFileExts";
          poA.lpData = (BYTE*) strHtmlFileExtsA;
          poA.dwData = (lstrlenA(strHtmlFileExtsA) + 1)*sizeof(char);
          poA.dwType = PO_STRING;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
          lstrcpyA( (LPSTR) strHtmlFileExtsW_0, strHtmlFileExtsA );
        }
        if (nStrCmpComment1Exts != 0)
        {
          poA.pOptionName = "Comment1FileExts";
          poA.lpData = (BYTE*) strComment1FileExtsA;
          poA.dwData = (lstrlenA(strComment1FileExtsA) + 1)*sizeof(char);
          poA.dwType = PO_STRING;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
          lstrcpyA( (LPSTR) strComment1FileExtsW_0, strComment1FileExtsA );
        }
        if (nStrCmpNextCharOk != 0)
        {
          poA.pOptionName = "next_char_ok";
          poA.lpData = (BYTE*) opt_szNextCharOkA;
          poA.dwData = (lstrlenA(opt_szNextCharOkA) + 1)*sizeof(char);
          poA.dwType = PO_STRING;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
          lstrcpyA( (LPSTR) opt_szNextCharOkW_0, opt_szNextCharOkA );
        }
        if (nStrCmpPrevCharOk != 0)
        {
          poA.pOptionName = "prev_char_ok";
          poA.lpData = (BYTE*) opt_szPrevCharOkA;
          poA.dwData = (lstrlenA(opt_szPrevCharOkA) + 1)*sizeof(char);
          poA.dwType = PO_STRING;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
          lstrcpyA( (LPSTR) opt_szPrevCharOkW_0, opt_szPrevCharOkA );
        }
        if (nStrCmpUsrBrPairs != 0)
        {
          poA.pOptionName = "user_brpairs";
          poA.lpData = (BYTE*) opt_szUserBracketsA;
          poA.dwData = (lstrlenA(opt_szUserBracketsA) + 1)*sizeof(char);
          poA.dwType = PO_STRING;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
          lstrcpyA( (LPSTR) opt_szUserBracketsW_0, opt_szUserBracketsA );
        }
        for (i = 0; i < OPT_DWORD_COUNT; i++)
        {
          if (g_dwOptions[i] != g_dwOptions0[i])
          {
            poA.pOptionName = (char*) cszOptDwordNames[i];
            poA.lpData = (BYTE*) &g_dwOptions[i];
            poA.dwData = sizeof(DWORD);
            poA.dwType = PO_DWORD;
            SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA);
            g_dwOptions0[i] = g_dwOptions[i];
          }
        }
        SendMessage(g_hMainWnd, AKD_ENDOPTIONS, (WPARAM) hOptions, 0);
      }

    }
    else
    {
      HANDLE        hOptions;
      PLUGINOPTIONW poW;

      hOptions = (HANDLE) SendMessage(g_hMainWnd,
        AKD_BEGINOPTIONS, POB_SAVE, (LPARAM) L"XBrackets");

      if (hOptions)
      {
        if (dwNewOptionsFlags != opt_dwOptionsFlags0)
        {
          poW.pOptionName = L"OptionsFlags";
          poW.lpData = (BYTE*) &dwNewOptionsFlags;
          poW.dwData = sizeof(DWORD);
          poW.dwType = PO_DWORD;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
          opt_dwOptionsFlags0 = dwNewOptionsFlags;
        }
        if (dwNewHighlightRGB != opt_dwHighlightRGB0)
        {
          poW.pOptionName = L"HighlightRGB";
          poW.lpData = (BYTE*) &dwNewHighlightRGB;
          poW.dwData = sizeof(DWORD);
          poW.dwType = PO_BINARY;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
          opt_dwHighlightRGB0 = dwNewHighlightRGB;
        }
        if (nStrCmpHtmlExts != 0)
        {
          poW.pOptionName = L"HtmlFileExts";
          poW.lpData = (BYTE*) strHtmlFileExtsW;
          poW.dwData = (lstrlenW(strHtmlFileExtsW) + 1)*sizeof(wchar_t);
          poW.dwType = PO_STRING;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
          lstrcpyW(strHtmlFileExtsW_0, strHtmlFileExtsW);
        }
        if (nStrCmpComment1Exts != 0)
        {
          poW.pOptionName = L"Comment1FileExts";
          poW.lpData = (BYTE*) strComment1FileExtsW;
          poW.dwData = (lstrlenW(strComment1FileExtsW) + 1)*sizeof(wchar_t);
          poW.dwType = PO_STRING;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
          lstrcpyW(strComment1FileExtsW_0, strComment1FileExtsW);
        }
        if (nStrCmpNextCharOk != 0)
        {
          poW.pOptionName = L"next_char_ok";
          poW.lpData = (BYTE*) opt_szNextCharOkW;
          poW.dwData = (lstrlenW(opt_szNextCharOkW) + 1)*sizeof(wchar_t);
          poW.dwType = PO_STRING;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
          lstrcpyW(opt_szNextCharOkW_0, opt_szNextCharOkW);
        }
        if (nStrCmpPrevCharOk != 0)
        {
          poW.pOptionName = L"prev_char_ok";
          poW.lpData = (BYTE*) opt_szPrevCharOkW;
          poW.dwData = (lstrlenW(opt_szPrevCharOkW) + 1)*sizeof(wchar_t);
          poW.dwType = PO_STRING;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
          lstrcpyW(opt_szPrevCharOkW_0, opt_szPrevCharOkW);
        }
        if (nStrCmpUsrBrPairs != 0)
        {
          poW.pOptionName = L"user_brpairs";
          poW.lpData = (BYTE*) opt_szUserBracketsW;
          poW.dwData = (lstrlenW(opt_szUserBracketsW) + 1)*sizeof(wchar_t);
          poW.dwType = PO_STRING;
          SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
          lstrcpyW(opt_szUserBracketsW_0, opt_szUserBracketsW);
        }
        for (i = 0; i < OPT_DWORD_COUNT; i++)
        {
          if (g_dwOptions[i] != g_dwOptions0[i])
          {
            poW.pOptionName = (wchar_t*) cwszOptDwordNames[i];
            poW.lpData = (BYTE*) &g_dwOptions[i];
            poW.dwData = sizeof(DWORD);
            poW.dwType = PO_DWORD;
            SendMessage(g_hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW);
            g_dwOptions0[i] = g_dwOptions[i];
          }
        }
        SendMessage(g_hMainWnd, AKD_ENDOPTIONS, (WPARAM) hOptions, 0);
      }

    }
  }

}

// Entry point
/*extern "C"*/
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
  }
  return TRUE;
}
