#include "SettingsDlg.h"
#include "AnyWindow.h"
#include "resource.h"
#include "c_base/HexStr.h"
#include "Plugin.h"
#include "XBracketsLng.h"

extern HWND        g_hMainWnd;
extern BOOL        g_bOldWindows;
extern BOOL        g_bInitialized;
extern LANGID      g_LangSystem;

extern BOOL        bBracketsAutoComplete;
extern BOOL        bBracketsHighlight;
extern BOOL        bBracketsHighlightVisibleArea;
extern BOOL        bBracketsRightExistsOK;
extern BOOL        bBracketsDoSingleQuote;
extern BOOL        bBracketsDoTag;
extern BOOL        bBracketsDoTag2;
extern BOOL        bBracketsDoTagIf;
extern BOOL        bBracketsSkipEscaped;
extern BOOL        bBracketsSkipComment1;
extern COLORREF    bracketsColourHighlight;
extern char        strBracketsColourA[STR_COLOUR_SIZE];
extern wchar_t     strBracketsColourW[STR_COLOUR_SIZE];
extern char        strHtmlFileExtsA[STR_FILEEXTS_SIZE];
extern wchar_t     strHtmlFileExtsW[STR_FILEEXTS_SIZE];
extern char        strComment1FileExtsA[STR_FILEEXTS_SIZE];
extern wchar_t     strComment1FileExtsW[STR_FILEEXTS_SIZE];
extern wchar_t     strPluginFuncMainW[STR_PLUGINFUNC_SIZE];

BOOL CheckBox_IsChecked(HWND hDlg, UINT idCheckBox);
BOOL CheckBox_SetCheck(HWND hDlg, UINT idCheckBox, BOOL bChecked);
BOOL DlgItem_EnableWindow(HWND hDlg, UINT idDlgItem, BOOL bEnable);
BOOL DlgItem_SetText(HWND hDlg, UINT idDlgItem, BOOL bUnicode, const void* pszText);

BOOL SettingsDlg_OnOK(HWND hDlg);
void SettingsDlg_OnChBracketsAutoCompleteClicked(HWND hDlg);
void SettingsDlg_OnChBracketsHighlightClicked(HWND hDlg);
void SettingsDlg_OnChBracketsDoTagClicked(HWND hDlg);
void SettingsDlg_OnChBracketsDoTagIfClicked(HWND hDlg);
void SettingsDlg_OnChBracketsSkipComment1(HWND hDlg);
void SettingsDlg_OnStPluginStateDblClicked(HWND hDlg);
void SettingsDlg_OnInitDialog(HWND hDlg);

BOOL /*INT_PTR*/ CALLBACK SettingsDlgProc(
  HWND   hDlg, 
  UINT   uMessage, 
  WPARAM wParam, 
  LPARAM lParam)
{
  if (uMessage == WM_COMMAND)
  {
    switch (LOWORD(wParam)) 
    {
      case IDC_CH_BRACKETS_HIGHLIGHT:
      {
        if (HIWORD(wParam) == BN_CLICKED)
        {
          SettingsDlg_OnChBracketsHighlightClicked(hDlg);
        }
        break;
      }
      case IDC_CH_BRACKETS_AUTOCOMPLETE:
      {
        if (HIWORD(wParam) == BN_CLICKED)
        {
          SettingsDlg_OnChBracketsAutoCompleteClicked(hDlg);
        }
        break;
      }
      case IDC_CH_BRACKETS_DOTAG:
      {
        if (HIWORD(wParam) == BN_CLICKED)
        {
          SettingsDlg_OnChBracketsDoTagClicked(hDlg);
        }
        break;
      }
      case IDC_CH_BRACKETS_DOTAGIF:
      {
        if (HIWORD(wParam) == BN_CLICKED)
        {
          SettingsDlg_OnChBracketsDoTagIfClicked(hDlg);
        }
        break;
      }
      case IDC_CH_BRACKETS_SKIPCOMMENT1:
      {
        if (HIWORD(wParam) == BN_CLICKED)
        {
          SettingsDlg_OnChBracketsSkipComment1(hDlg);
        }
        break;
      }
      case IDC_ST_PLUGINSTATE:
      {
        if (HIWORD(wParam) == STN_DBLCLK)
        {
          SettingsDlg_OnStPluginStateDblClicked(hDlg);
        }
        break;
      }
      case IDC_BT_OK: // Run
      case IDOK:
      {
        if (SettingsDlg_OnOK(hDlg))
        {
          EndDialog(hDlg, 1); // OK - returns 1
        }
        return 1;
      }
      case IDC_BT_CANCEL: // Cancel
      case IDCANCEL:
      {
        EndDialog(hDlg, 0); // Cancel - returns 0
        return 1;
      }
      default:
        break;
    }
  }

  else if (uMessage == WM_NOTIFY)
  {
    
  }

  else if (uMessage == WM_INITDIALOG)
  {
    SettingsDlg_OnInitDialog(hDlg);
  }

  return 0;
}

BOOL SettingsDlg_OnOK(HWND hDlg)
{
  HWND hEd = GetDlgItem(hDlg, IDC_ED_COLOR);
  
  if (hEd)
  {
    /*c_base::*/byte_t btColour[4];
    
    if (g_bOldWindows)
    {
      GetWindowTextA(hEd, strBracketsColourA, STR_COLOUR_SIZE-1);
      if (/*c_base::*/hexstr2buf(strBracketsColourA, btColour, 3) != 3)
      {
        MessageBoxA(hDlg, "Wrong HEX value of the Colour", 
          "XBrackets Error", MB_OK | MB_ICONERROR);
        return FALSE;
      }
    }
    else
    {
      GetWindowTextW(hEd, strBracketsColourW, STR_COLOUR_SIZE-1);
      if (/*c_base::*/hexstrw2buf(strBracketsColourW, btColour, 3) != 3)
      {
        MessageBoxW(hDlg, L"Wrong HEX value of the Colour", 
          L"XBrackets Error", MB_OK | MB_ICONERROR);
        return FALSE;
      }
    }
    bracketsColourHighlight = RGB(btColour[0], btColour[1], btColour[2]);
  }

  hEd = GetDlgItem(hDlg, IDC_ED_BRACKETS_DOTAGIF);
  if (hEd)
  {
    if (g_bOldWindows)
      GetWindowTextA(hEd, strHtmlFileExtsA, STR_FILEEXTS_SIZE - 1);
    else
      GetWindowTextW(hEd, strHtmlFileExtsW, STR_FILEEXTS_SIZE - 1);
  }

  hEd = GetDlgItem(hDlg, IDC_ED_BRACKETS_SKIPCOMMENT1);
  if (hEd)
  {
    if (g_bOldWindows)
      GetWindowTextA(hEd, strComment1FileExtsA, STR_FILEEXTS_SIZE - 1);
    else
      GetWindowTextW(hEd, strComment1FileExtsW, STR_FILEEXTS_SIZE - 1);
  }

  bBracketsAutoComplete = 
    CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_AUTOCOMPLETE);
  bBracketsHighlight = 
    CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_HIGHLIGHT);
  bBracketsHighlightVisibleArea =
    CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_HIGHLIGHT_VISIBLEAREA);
  bBracketsRightExistsOK = 
    CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_RIGHTEXISTS_OK);
  bBracketsDoSingleQuote = 
    CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_DOSINGLEQUOTE);
  bBracketsDoTag = 
    CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_DOTAG);
  bBracketsDoTag2 = 
    CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_DOTAG2);
  bBracketsDoTagIf = 
    CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_DOTAGIF);
  bBracketsSkipEscaped =
    CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_SKIPESCAPED);
  bBracketsSkipComment1 = 
    CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_SKIPCOMMENT1);
  
  return TRUE;
}

void SettingsDlg_OnChBracketsAutoCompleteClicked(HWND hDlg)
{
  BOOL bEnable = CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_AUTOCOMPLETE);
  DlgItem_EnableWindow(hDlg, IDC_CH_BRACKETS_RIGHTEXISTS_OK, bEnable);
}

void SettingsDlg_OnChBracketsHighlightClicked(HWND hDlg)
{
  BOOL bEnable = CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_HIGHLIGHT);
  DlgItem_EnableWindow(hDlg, IDC_CH_BRACKETS_HIGHLIGHT_VISIBLEAREA, bEnable);
  DlgItem_EnableWindow(hDlg, IDC_ST_COLOR, bEnable);
  DlgItem_EnableWindow(hDlg, IDC_ED_COLOR, bEnable);
}

void SettingsDlg_OnChBracketsDoTagClicked(HWND hDlg)
{
  BOOL bEnable = CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_DOTAG);
  DlgItem_EnableWindow(hDlg, IDC_CH_BRACKETS_DOTAG2, bEnable);
  DlgItem_EnableWindow(hDlg, IDC_CH_BRACKETS_DOTAGIF, bEnable);
  if (bEnable)
  {
    bEnable = CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_DOTAGIF);
  }
  DlgItem_EnableWindow(hDlg, IDC_ED_BRACKETS_DOTAGIF, bEnable);
}

void SettingsDlg_OnChBracketsDoTagIfClicked(HWND hDlg)
{
  BOOL bEnable = CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_DOTAGIF);
  DlgItem_EnableWindow(hDlg, IDC_ED_BRACKETS_DOTAGIF, bEnable);
}

void SettingsDlg_OnChBracketsSkipComment1(HWND hDlg)
{
  BOOL bEnable = CheckBox_IsChecked(hDlg, IDC_CH_BRACKETS_SKIPCOMMENT1);
  DlgItem_EnableWindow(hDlg, IDC_ED_BRACKETS_SKIPCOMMENT1, bEnable);
  DlgItem_EnableWindow(hDlg, IDC_ST_BRACKETS_SKIPCOMMENT1, bEnable);
}

void showPluginStatus(HWND hDlg)
{
  if (g_bOldWindows)
  {
    DlgItem_SetText(hDlg, IDC_ST_PLUGINSTATE, FALSE,
      g_bInitialized ? xbrGetStrA(XBR_STR_PLUGINSTATUSACTIVE, g_LangSystem) : 
        xbrGetStrA(XBR_STR_PLUGINSTATUSNOTACTIVE, g_LangSystem) );
  }
  else
  {
    DlgItem_SetText(hDlg, IDC_ST_PLUGINSTATE, TRUE,
      g_bInitialized ? xbrGetStrW(XBR_STR_PLUGINSTATUSACTIVE, g_LangSystem) :
        xbrGetStrW(XBR_STR_PLUGINSTATUSNOTACTIVE, g_LangSystem) );
  }
}

void SettingsDlg_OnStPluginStateDblClicked(HWND hDlg)
{
  BOOL bDoInitializeHere = TRUE;
  
  if (g_bInitialized)
  {
    Uninitialize(FALSE); 
    bDoInitializeHere = FALSE;
  }
  
  if (g_bOldWindows)
  {
    if (bDoInitializeHere)
    {
      // XBrackets::Main is unloaded, let's load it
      PLUGINCALLSENDA pcsA;
      pcsA.pFunction = (char *) strPluginFuncMainW;
      pcsA.bOnStart = FALSE;
      pcsA.lParam = 0;
      pcsA.lpbAutoLoad=NULL;
      SendMessage( g_hMainWnd, AKD_DLLCALL, 0, (LPARAM) &pcsA );
    }
    else
    {
      PLUGINFUNCTIONA* pfA = (PLUGINFUNCTIONA *) SendMessage( g_hMainWnd, 
        AKD_DLLFIND, (WPARAM) strPluginFuncMainW, 0 );
      if (pfA) 
      { 
        pfA->bRunning = FALSE; 
      }
    }
  }
  else
  {
    if (bDoInitializeHere)
    {
      // XBrackets::Main is unloaded, let's load it
      PLUGINCALLSENDW pcsW;
      pcsW.pFunction = (wchar_t *) strPluginFuncMainW;
      pcsW.bOnStart = FALSE;
      pcsW.lParam = 0;
      pcsW.lpbAutoLoad=NULL;
      SendMessage( g_hMainWnd, AKD_DLLCALL, 0, (LPARAM) &pcsW );
    }
    else
    {
      PLUGINFUNCTIONW* pfW = (PLUGINFUNCTIONW *) SendMessage( g_hMainWnd,
        AKD_DLLFIND, (WPARAM) strPluginFuncMainW, 0 );
      if (pfW)
      {
        pfW->bRunning = FALSE;
      } 
    }
  }
 
  showPluginStatus(hDlg);
}

void SettingsDlg_OnInitDialog(HWND hDlg)
{
  HWND       hEd;

  xbrSetSettingsDlgLang(hDlg, g_LangSystem);

  AnyWindow_CenterWindow(hDlg, g_hMainWnd, FALSE);

  CheckBox_SetCheck(hDlg, 
    IDC_CH_BRACKETS_AUTOCOMPLETE, bBracketsAutoComplete);
  CheckBox_SetCheck(hDlg, 
    IDC_CH_BRACKETS_RIGHTEXISTS_OK, bBracketsRightExistsOK);
  CheckBox_SetCheck(hDlg, 
    IDC_CH_BRACKETS_HIGHLIGHT, bBracketsHighlight);
  CheckBox_SetCheck(hDlg, 
    IDC_CH_BRACKETS_HIGHLIGHT_VISIBLEAREA, bBracketsHighlightVisibleArea);
  CheckBox_SetCheck(hDlg, 
    IDC_CH_BRACKETS_DOSINGLEQUOTE, bBracketsDoSingleQuote);
  CheckBox_SetCheck(hDlg, 
    IDC_CH_BRACKETS_DOTAG, bBracketsDoTag);
  CheckBox_SetCheck(hDlg, 
    IDC_CH_BRACKETS_DOTAG2, bBracketsDoTag2);
  CheckBox_SetCheck(hDlg, 
    IDC_CH_BRACKETS_DOTAGIF, bBracketsDoTagIf);
  CheckBox_SetCheck(hDlg,
    IDC_CH_BRACKETS_SKIPESCAPED, bBracketsSkipEscaped);
  CheckBox_SetCheck(hDlg, 
    IDC_CH_BRACKETS_SKIPCOMMENT1, bBracketsSkipComment1);
  DlgItem_EnableWindow(hDlg, 
    IDC_CH_BRACKETS_RIGHTEXISTS_OK, bBracketsAutoComplete);
  DlgItem_EnableWindow(hDlg, 
    IDC_CH_BRACKETS_HIGHLIGHT_VISIBLEAREA, bBracketsHighlight);
  DlgItem_EnableWindow(hDlg, IDC_ST_COLOR, bBracketsHighlight);
  DlgItem_EnableWindow(hDlg, IDC_ED_COLOR, bBracketsHighlight);
  DlgItem_EnableWindow(hDlg, 
    IDC_ED_BRACKETS_SKIPCOMMENT1, bBracketsSkipComment1);
  DlgItem_EnableWindow(hDlg, 
    IDC_ST_BRACKETS_SKIPCOMMENT1, bBracketsSkipComment1);

  DlgItem_EnableWindow(hDlg, IDC_CH_BRACKETS_DOTAG2, bBracketsDoTag);
  DlgItem_EnableWindow(hDlg, IDC_CH_BRACKETS_DOTAGIF, bBracketsDoTag);
  DlgItem_EnableWindow(hDlg, IDC_ED_BRACKETS_DOTAGIF, 
      bBracketsDoTag ? bBracketsDoTagIf : FALSE);

  showPluginStatus(hDlg);
  
  hEd = GetDlgItem(hDlg, IDC_ED_BRACKETS_DOTAGIF);
  if (hEd)
  {
    SendMessage(hEd, EM_LIMITTEXT, (WPARAM) (STR_FILEEXTS_SIZE - 1), 0);
      
    if (g_bOldWindows)
      SetWindowTextA(hEd, strHtmlFileExtsA);
    else
      SetWindowTextW(hEd, strHtmlFileExtsW);
  }

  hEd = GetDlgItem(hDlg, IDC_ED_BRACKETS_SKIPCOMMENT1);
  if (hEd)
  {
    SendMessage(hEd, EM_LIMITTEXT, (WPARAM) (STR_FILEEXTS_SIZE - 1), 0);
      
    if (g_bOldWindows)
      SetWindowTextA(hEd, strComment1FileExtsA);
    else
      SetWindowTextW(hEd, strComment1FileExtsW);
  }

  hEd = GetDlgItem(hDlg, IDC_ED_COLOR);
  if (hEd)
  {
    /*c_base::*/byte_t btColour[4];

    SendMessage(hEd, EM_LIMITTEXT, (WPARAM) 10, 0);
    
    btColour[0] = GetRValue(bracketsColourHighlight);
    btColour[1] = GetGValue(bracketsColourHighlight);
    btColour[2] = GetBValue(bracketsColourHighlight);
    if (g_bOldWindows)
    {
      /*c_base::*/buf2hexstr(btColour, 3, strBracketsColourA, STR_COLOUR_SIZE, " ");
      SetWindowTextA(hEd, strBracketsColourA);
    }
    else
    {
      /*c_base::*/buf2hexstrw(btColour, 3, strBracketsColourW, STR_COLOUR_SIZE, L" ");
      SetWindowTextW(hEd, strBracketsColourW);
    }
  }
 
}

//---------------------------------------------------------------------------

BOOL CheckBox_IsChecked(HWND hDlg, UINT idCheckBox)
{
  HWND hCheckBox = GetDlgItem(hDlg, idCheckBox);
  if (hCheckBox)
  {
    return ((SendMessage(hCheckBox, BM_GETCHECK, 
               0, 0) == BST_CHECKED) ? TRUE : FALSE);
  }
  return FALSE;
}

BOOL CheckBox_SetCheck(HWND hDlg, UINT idCheckBox, BOOL bChecked)
{
  HWND hCheckBox = GetDlgItem(hDlg, idCheckBox);
  if (hCheckBox)
  {
    SendMessage(hCheckBox, BM_SETCHECK,
      (WPARAM) (bChecked ? BST_CHECKED : BST_UNCHECKED), 0);
    return TRUE;
  }
  return FALSE;
}

BOOL DlgItem_EnableWindow(HWND hDlg, UINT idDlgItem, BOOL bEnable)
{
  HWND hDlgItem = GetDlgItem(hDlg, idDlgItem);
  if (hDlgItem)
  {
    return EnableWindow(hDlgItem, bEnable);
  }
  return FALSE;
}

BOOL DlgItem_SetText(HWND hDlg, UINT idDlgItem, BOOL bUnicode, const void* pszText)
{
  HWND hDlgItem = GetDlgItem(hDlg, idDlgItem);
  if (hDlgItem)
  {
    if (!bUnicode)
      return SetWindowTextA(hDlgItem, (const char *) pszText);
    else
      return SetWindowTextW(hDlgItem, (const wchar_t *) pszText);
  }
  return FALSE;
}
